/* advautoage_h_gm2d_divide_seg_by_seg.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

#define DEFAULT_EPS 0.00001

double Eps = DEFAULT_EPS;



int Equal (double value0, double value1) 
{
  return fabs (value1 - value0) < Eps;
}

int Between (double value, double range0, double range1) 
{
  return (range0 - Eps < value && value < range1 + Eps)
    || (range1 - Eps < value && value < range0 + Eps);
}

double Distance (double x0, double y0, 
		 double x1, double y1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) + (y1 - y0) * (y1 - y0));
}

int PointsAreSame (double x0, double y0, 
		   double x1, double y1) 
{
  return Distance (x0, y0, 
		   x1, y1) < Eps;
}

int SegsAreSame 
(double seg0_x0, double seg0_y0, 
 double seg0_x1, double seg0_y1, 
 double seg1_x0, double seg1_y0, 
 double seg1_x1, double seg1_y1) 
{
  if (PointsAreSame (seg0_x0, seg0_y0,
		     seg1_x0, seg1_y0)
      && PointsAreSame (seg0_x1, seg0_y1,
			seg1_x1, seg1_y1)) {
    return 1;
  }
  if (PointsAreSame (seg0_x0, seg0_y0,
		     seg1_x1, seg1_y1)
      && PointsAreSame (seg0_x1, seg0_y1,
			seg1_x0, seg1_y0)) {
    return 1;
  }
  return 0;
}

int SegHasPointOnEnd
(double x0, double y0,
 double x1, double y1,
 double x, double y)
{
  return PointsAreSame (x, y,
			x0, y0)
    || PointsAreSame (x, y,
		      x1, y1);
}

void GetLineEquation 
(double x0, double y0, 
 double x1, double y1, 
 double *a_OUT, double *b_OUT, double *c_OUT) 
{
  double dx = x1 - x0;
  double dy = y1 - y0;

  double a0 = -dy;
  double b0 = dx;
  double c0 = dy * x0 - y0 * dx;

  double len = sqrt (a0 * a0 + b0 * b0);

  *a_OUT = a0 / len;  
  *b_OUT = b0 / len;  
  *c_OUT = c0 / len;
}

int SolveLinearEquation
(double a00, double a01,
 double a10, double a11, 
 double b0, double b1,
 double *x_OUT, double *y_OUT)
{
  double det = a00 * a11 - a01 * a10;

  /* a00 x + a01 y + b0 = 0 */
  /* a10 x + a11 y + b1 = 0 */

  if (fabs (det) < Eps * Eps) return 0;

  *x_OUT = (a01 * b1 - a11 * b0) / det;
  *y_OUT = (a10 * b0 - a00 * b1) / det;
  return 1;
}

int SegContainsPoint
(double x0, double y0,
 double x1, double y1,
 double x, double y)
{
  double a, b, c;
  
  GetLineEquation 
    (x0, y0,
     x1, y1,
     &a, &b, &c);
      
  if (fabs (a * x + b * y + c) < Eps) {
    /* point is on line of segment */

    if (Equal (x0, x1)) {
      assert(!Equal (y0, y1));
      
      return Between (y, 
		      y0, y1);
    } else {
      return Between (x, 
		      x0, x1);
    }
  } else {
    return 0;
  }
}

int SegJustContainsPoint
(double x0, double y0,
 double x1, double y1,
 double x, double y)
{
  return SegContainsPoint
    (x0, y0,
     x1, y1,
     x, y)
    && !SegHasPointOnEnd
    (x0, y0,
     x1, y1,
     x, y);
}

int GetIntersectionBetweenSegs 
(double seg0_x0, double seg0_y0, 
 double seg0_x1, double seg0_y1, 
 double seg1_x0, double seg1_y0, 
 double seg1_x1, double seg1_y1,
 double *intX_OUT, double *intY_OUT) 
{
  double a0, b0, c0;
  double a1, b1, c1;
  int flag;
  
  GetLineEquation (seg0_x0, seg0_y0, 
		   seg0_x1, seg0_y1, 
		   &a0, &b0, &c0);
  GetLineEquation (seg1_x0, seg1_y0, 
		   seg1_x1, seg1_y1, 
		   &a1, &b1, &c1);
  flag = SolveLinearEquation 
    (a0, b0, 
     a1, b1, 
     c0, c1,
     intX_OUT, intY_OUT);
  if (flag) {
    if (Between (*intX_OUT, seg0_x0, seg0_x1)
	&& Between (*intY_OUT, seg0_y0, seg0_y1)) {
      if (Between (*intX_OUT, seg1_x0, seg1_x1)
	  && Between (*intY_OUT, seg1_y0, seg1_y1)) {
	return 1;
      }
    }
  }
  return 0;
}



/************************
        Segments
*************************/

#define MAX_SEGS 100000

/* Segment */

int NSegs;
double Seg_coord0[MAX_SEGS][2];
double Seg_coord1[MAX_SEGS][2];



/* By-Segment */

int NBySegs;
double BySeg_coord0[MAX_SEGS][2];
double BySeg_coord1[MAX_SEGS][2];



/* Out-Segment */

int NOutSegs;
double OutSeg_coord0[MAX_SEGS][2];
double OutSeg_coord1[MAX_SEGS][2];



void ReadSegFile (const char *fileName)
{
  FILE *fp;
  int iSeg;

  assert(NSegs == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NSegs);
  assert(0 <= NSegs);
  assert(NSegs <= MAX_SEGS);
  
  for (iSeg = 0; iSeg < NSegs; iSeg++) {
    int segId;
    double x0, y0, x1, y1;

    fscanf (fp, "%d %lf %lf %lf %lf", 
	    &segId,
	    &x0, &y0, 
	    &x1, &y1);
    assert(segId == iSeg);

    Seg_coord0[iSeg][0] = x0;
    Seg_coord0[iSeg][1] = y0; 

    Seg_coord1[iSeg][0] = x1;
    Seg_coord1[iSeg][1] = y1; 
  }

  fclose (fp);
}

void ReadBySegFile (const char *fileName)
{
  FILE *fp;
  int iSeg;

  assert(NBySegs == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NBySegs);
  assert(0 <= NBySegs);
  assert(NBySegs <= MAX_SEGS);
  
  for (iSeg = 0; iSeg < NBySegs; iSeg++) {
    int segId;
    double x0, y0, x1, y1;

    fscanf (fp, "%d %lf %lf %lf %lf", 
	    &segId,
	    &x0, &y0, 
	    &x1, &y1);
    assert(segId == iSeg);

    BySeg_coord0[iSeg][0] = x0;
    BySeg_coord0[iSeg][1] = y0; 

    BySeg_coord1[iSeg][0] = x1;
    BySeg_coord1[iSeg][1] = y1; 
  }

  fclose (fp);
}

void AddOutSeg 
(double x0, double y0,
 double x1, double y1)
{
  assert(NOutSegs < MAX_SEGS);

  OutSeg_coord0[NOutSegs][0] = x0;
  OutSeg_coord0[NOutSegs][1] = y0;
  
  OutSeg_coord1[NOutSegs][0] = x1;
  OutSeg_coord1[NOutSegs][1] = y1;

  NOutSegs++;
}

void WriteOutSegFile (const char *fileName)
{
  FILE *fp;
  int iSeg;

  assert(0 < NOutSegs);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  fprintf (fp, "%d\n", 
	   NOutSegs);
  
  for (iSeg = 0; iSeg < NOutSegs; iSeg++) {
    double x0 = OutSeg_coord0[iSeg][0];
    double y0 = OutSeg_coord0[iSeg][1];

    double x1 = OutSeg_coord1[iSeg][0];
    double y1 = OutSeg_coord1[iSeg][1];

    fprintf (fp, "%d    %22.15e %22.15e   %22.15e %22.15e\n", 
	     iSeg,
	     x0, y0,
	     x1, y1);
  }

  fclose (fp);
}



/************************
	 Command
*************************/

/* Tmp-Segment */

int NTmpSegs;
double TmpSeg_coord0[MAX_SEGS][2];
double TmpSeg_coord1[MAX_SEGS][2];

static void ClearTmpSeg (void)
{
  NTmpSegs = 0;
}

static void AddTmpSeg 
(double x0, double y0,
 double x1, double y1)
{
  assert(NTmpSegs < MAX_SEGS);

  TmpSeg_coord0[NTmpSegs][0] = x0;
  TmpSeg_coord0[NTmpSegs][1] = y0;
  
  TmpSeg_coord1[NTmpSegs][0] = x1;
  TmpSeg_coord1[NTmpSegs][1] = y1;

  NTmpSegs++;
}

static void RemoveTmpSeg 
(double x0, double y0,
 double x1, double y1)
{
  int iSeg;
  int sameSegId;

  sameSegId = -1;
  for (iSeg = 0; iSeg < NTmpSegs; iSeg++) {
    double tx0 = TmpSeg_coord0[iSeg][0];
    double ty0 = TmpSeg_coord0[iSeg][1];

    double tx1 = TmpSeg_coord1[iSeg][0];
    double ty1 = TmpSeg_coord1[iSeg][1];

    if (SegsAreSame (tx0, ty0,
		     tx1, ty1,
		     x0, y0,
		     x1, y1)) {
      sameSegId = iSeg;
      break;
    }
  }

  if (0 <= sameSegId) {
    for (iSeg = sameSegId; iSeg < NTmpSegs - 1; iSeg++) {
      TmpSeg_coord0[iSeg][0] = TmpSeg_coord0[iSeg + 1][0];
      TmpSeg_coord0[iSeg][1] = TmpSeg_coord0[iSeg + 1][1];
      
      TmpSeg_coord1[iSeg][0] = TmpSeg_coord1[iSeg + 1][0];
      TmpSeg_coord1[iSeg][1] = TmpSeg_coord1[iSeg + 1][1];
    }
    NTmpSegs--;
  }
}

static void DivideBy 
(double x0, double y0,
 double x1, double y1)
{
  int iSeg;
  int intersectionIsFound;

  for (;;) {
    intersectionIsFound = 0;
    for (iSeg = 0; iSeg < NTmpSegs; iSeg++) {
      double tx0 = TmpSeg_coord0[iSeg][0];
      double ty0 = TmpSeg_coord0[iSeg][1];
      
      double tx1 = TmpSeg_coord1[iSeg][0];
      double ty1 = TmpSeg_coord1[iSeg][1];
      
      int flag;
      double intX, intY;
      
      flag = 0;
      if (SegJustContainsPoint
	  (tx0, ty0,
	   tx1, ty1,
	   x0, y0)) {
	flag = 1;
	intX = x0;
	intY = y0;
      } else if (SegJustContainsPoint
		 (tx0, ty0,
		  tx1, ty1,
		  x1, y1)) {
	flag = 1;
	intX = x1;
	intY = y1;
      } else {
	flag = GetIntersectionBetweenSegs
	  (tx0, ty0,
	   tx1, ty1,
	   x0, y0,
	   x1, y1,
	   &intX, &intY);
	if (flag) {
	  if (SegHasPointOnEnd (tx0, ty0,
				tx1, ty1,
				intX, intY)) {
	    flag = 0;
	  }
	}
      }

      if (flag) {
	RemoveTmpSeg (tx0, ty0,
		      tx1, ty1);
	AddTmpSeg (tx0, ty0,
		   intX, intY);
	AddTmpSeg (intX, intY,
		   tx1, ty1);
	intersectionIsFound = 1;
	break;
      }
    }

    if (!intersectionIsFound) {
      break;
    }
  }
}

static void MoveTmpSegToOut (void)
{
  int iSeg;

  for (iSeg = 0; iSeg < NTmpSegs; iSeg++) {
    double tx0 = TmpSeg_coord0[iSeg][0];
    double ty0 = TmpSeg_coord0[iSeg][1];
    
    double tx1 = TmpSeg_coord1[iSeg][0];
    double ty1 = TmpSeg_coord1[iSeg][1];
    
    AddOutSeg (tx0, ty0,
	       tx1, ty1);
  }
}

void PerformCommand (void)
{
  int iSeg, jSeg;
  
  for (iSeg = 0; iSeg < NSegs; iSeg++) {
    double x0 = Seg_coord0[iSeg][0];
    double y0 = Seg_coord0[iSeg][1];

    double x1 = Seg_coord1[iSeg][0];
    double y1 = Seg_coord1[iSeg][1];

    ClearTmpSeg ();
    AddTmpSeg (x0, y0, 
	       x1, y1);

    for (jSeg = 0; jSeg < NBySegs; jSeg++) {
      double bx0 = BySeg_coord0[jSeg][0];
      double by0 = BySeg_coord0[jSeg][1];
      
      double bx1 = BySeg_coord1[jSeg][0];
      double by1 = BySeg_coord1[jSeg][1];
      
      DivideBy (bx0, by0,
		bx1, by1);
    }

    MoveTmpSegToOut ();
  }
}



int main (int argc, char *argv[])
{
  if (argc != 4) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautoage_h_gm2d_divide_seg_by_seg  seg_file  by_seg_file  \n"
       "        seg_file_OUT \n");
    exit (1);
  }

  ReadSegFile (argv[1]);
  ReadBySegFile (argv[2]);

  PerformCommand ();

  WriteOutSegFile (argv[3]);

  return 0;
}
