/* advautoage_h_gm2d_seg2polygon.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/* NEED TO REDEFINE */
/* if multiple polygons are generated, ??? */



/************************
	 Basic
*************************/

#define DEFAULT_EPS 0.00001

double Eps = DEFAULT_EPS;



int Equal (double value0, double value1) 
{
  return fabs (value1 - value0) < Eps;
}

int Between (double value, double range0, double range1) 
{
  return (range0 - Eps < value && value < range1 + Eps)
    || (range1 - Eps < value && value < range0 + Eps);
}

double Distance (double x0, double y0, 
		 double x1, double y1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) + (y1 - y0) * (y1 - y0));
}

int PointsAreSame (double x0, double y0, 
		   double x1, double y1) 
{
  return Distance (x0, y0, 
		   x1, y1) < Eps;
}

void GetLineEquation 
(double x0, double y0, 
 double x1, double y1, 
 double *a_OUT, double *b_OUT, double *c_OUT) 
{
  double dx = x1 - x0;
  double dy = y1 - y0;

  double a0 = -dy;
  double b0 = dx;
  double c0 = dy * x0 - y0 * dx;

  double len = sqrt (a0 * a0 + b0 * b0);

  *a_OUT = a0 / len;  
  *b_OUT = b0 / len;  
  *c_OUT = c0 / len;
}

int SegContainsPoint
(double x0, double y0,
 double x1, double y1,
 double x, double y)
{
  double a, b, c;
  
  GetLineEquation 
    (x0, y0,
     x1, y1,
     &a, &b, &c);
      
  if (fabs (a * x + b * y + c) < Eps) {
    /* point is on line of segment */

    if (Equal (x0, x1)) {
      assert(!Equal (y0, y1));
      
      return Between (y, 
		      y0, y1);
    } else {
      return Between (x, 
		      x0, x1);
    }
  } else {
    return 0;
  }
}

void GetVectorProduct3D 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  *x_OUT = y0 * z1 - z0 * y1;
  *y_OUT = z0 * x1 - x0 * z1;
  *z_OUT = x0 * y1 - y0 * x1;
}



/************************
     In-Out Segments
*************************/

#define MAX_INOUT_SEGS 100000

static int InOut_nSegs;
static double InOut_seg_coord0[MAX_INOUT_SEGS][2];
static double InOut_seg_coord1[MAX_INOUT_SEGS][2];

#define MAX_INOUT_INTERSECTIONS 1000
static int InOut_nIntersections;
static double InOut_intersectionYs[MAX_INOUT_INTERSECTIONS];



static double InOut_containsPointOnXAxis 
(int segId, 
 double px, double py)
{
  double sx0 = InOut_seg_coord0[segId][0];
  double sx1 = InOut_seg_coord1[segId][0];
  
  return Between (px, 
		  sx0, sx1);
}

static double InOut_containsPoint
(int segId, 
 double px, double py)
{
  double sx0 = InOut_seg_coord0[segId][0];
  double sy0 = InOut_seg_coord0[segId][1];
  
  double sx1 = InOut_seg_coord1[segId][0];
  double sy1 = InOut_seg_coord1[segId][1];
  
  return SegContainsPoint
    (sx0, sy0,
     sx1, sy1,
     px, py);
}

#define INOUT_RAY_UNKNOWN      0
#define INOUT_RAY_INTERSECT    1
#define INOUT_RAY_ON_SEGMENT  2

static int InOut_checkSegAndPointAlongYRay 
(int segId,
 double px, double py,
 double *intersectionY_OUT)
{
  int inside2D = InOut_containsPoint (segId,
				      px, py);
  
  if (inside2D) {
    return INOUT_RAY_ON_SEGMENT;

  } else {
    double sx0 = InOut_seg_coord0[segId][0];
    double sy0 = InOut_seg_coord0[segId][1];
    
    double sx1 = InOut_seg_coord1[segId][0];
    double sy1 = InOut_seg_coord1[segId][1];

    double a, b, c;

    GetLineEquation 
      (sx0, sy0,
       sx1, sy1,
       &a, &b, &c);
    
    if (fabs (b) < Eps) {
      /* parallel to Y axis */
      
      return INOUT_RAY_UNKNOWN;
    } else {
      int inside2D = InOut_containsPointOnXAxis (segId,
						 px, py);
      double signedDist = a * px + b * py + c;
      
      if (fabs (signedDist) < Eps) {
	/* point is on line of segment */
	/* on boundary, or unknown */
	
	if (inside2D) {
	  return INOUT_RAY_ON_SEGMENT;
	} else {
	  return INOUT_RAY_UNKNOWN;
	}
	
      } else {
	/* point is out of line of segment */
	/* intersect or unknown */
	
	if (inside2D) {
	  *intersectionY_OUT = - (c + a * px) / b;

	  if (py < *intersectionY_OUT) {
	    return INOUT_RAY_INTERSECT;
	  } else {
	    assert(*intersectionY_OUT < py);

	    return INOUT_RAY_UNKNOWN;
	  }
	} else {
	  return INOUT_RAY_UNKNOWN;
	}
	
      }
    }
  }

  assert(0);

  return INOUT_RAY_UNKNOWN;
}

static void InOut_clearIntersection (void)
{
  InOut_nIntersections = 0;
}

static void InOut_addIntersection (double y)
{
  int iIntersection;
  int found;
  
  found = 0;
  for (iIntersection = 0; 
       iIntersection < InOut_nIntersections; 
       iIntersection++) {
    if (fabs (InOut_intersectionYs[iIntersection] - y) < Eps) {
      found = 1;
      break;
    }
  }
  
  if (!found) {
    assert(InOut_nIntersections < MAX_INOUT_INTERSECTIONS);
    
    InOut_intersectionYs[InOut_nIntersections] = y;
    InOut_nIntersections++;
  }
}

void InOut_clear (void)
{
  InOut_nSegs = 0;
}

void InOut_addSeg 
(double x0, double y0,
 double x1, double y1)
{
  assert(InOut_nSegs < MAX_INOUT_SEGS);

  InOut_seg_coord0[InOut_nSegs][0] = x0;
  InOut_seg_coord0[InOut_nSegs][1] = y0;
  
  InOut_seg_coord1[InOut_nSegs][0] = x1;
  InOut_seg_coord1[InOut_nSegs][1] = y1;

  InOut_nSegs++;
}

#define INOUT_POINT_OUTSIDE      0
#define INOUT_POINT_INSIDE       1
#define INOUT_POINT_ON_BOUNDARY  2

int InOut_checkInout (double px, double py)
{  
  int iSeg;

  InOut_clearIntersection ();
  for (iSeg = 0; iSeg < InOut_nSegs; iSeg++) {
    double intersectionY;
    int checkResult = InOut_checkSegAndPointAlongYRay 
      (iSeg, 
       px, py,
       &intersectionY);

    switch (checkResult) {
    case INOUT_RAY_UNKNOWN:
      break;
    case INOUT_RAY_INTERSECT:
      InOut_addIntersection (intersectionY);
      break;
    case INOUT_RAY_ON_SEGMENT:
      return INOUT_POINT_ON_BOUNDARY;
    default:
      assert(0);
      break;
    }
  }

  if (InOut_nIntersections % 2 == 1) {
    return INOUT_POINT_INSIDE;
  } else {
    return INOUT_POINT_OUTSIDE;
  }
}



/************************
        Segments
*************************/

/* Segment */

#define MAX_SEGS 100000

int NSegs;
double Seg_coord0[MAX_SEGS][2];
double Seg_coord1[MAX_SEGS][2];
int Seg_hasMark[MAX_SEGS];



void ReadSegFile (const char *fileName)
{
  FILE *fp;
  int iSeg;

  assert(NSegs == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NSegs);
  assert(0 <= NSegs);
  assert(NSegs <= MAX_SEGS);
  
  for (iSeg = 0; iSeg < NSegs; iSeg++) {
    int segId;
    double x0, y0, x1, y1;

    fscanf (fp, "%d %lf %lf %lf %lf", 
	    &segId,
	    &x0, &y0, 
	    &x1, &y1);
    assert(segId == iSeg);

    Seg_coord0[iSeg][0] = x0;
    Seg_coord0[iSeg][1] = y0; 

    Seg_coord1[iSeg][0] = x1;
    Seg_coord1[iSeg][1] = y1; 

    Seg_hasMark[iSeg] = 0; 
  }

  fclose (fp);
}

void MarkSeg (int segId)
{
  assert(0 <= segId);
  assert(segId < NSegs);

  Seg_hasMark[segId] = 1;
}

int GetUnmarkedSeg (void)
{
  int iSeg;
  
  for (iSeg = 0; iSeg < NSegs; iSeg++) {
    if (!Seg_hasMark[iSeg]) {
      return iSeg;
    }
  }

  return -1;
}



/************************
	 Polygons
*************************/

#define MAX_POLYGONS 1000
#define MAX_POLYGON_LOOPS 100
#define MAX_POLYGON_VERTEXS 100

int NPolygons;
int Polygon_nLoops[MAX_POLYGONS];
int Polygon_nVertexs[MAX_POLYGONS][MAX_POLYGON_LOOPS];
double Polygon_vertexs[MAX_POLYGONS][MAX_POLYGON_LOOPS][MAX_POLYGON_VERTEXS][2];



int AddNewPolygon (void)
{
  int polygonId;

  polygonId = NPolygons;

  assert(NPolygons < MAX_POLYGONS);
  NPolygons++;

  return polygonId;
}

int AddNewLoopInPolygon (int polygonId)
{
  int loopId;

  assert(0 <= polygonId);
  assert(polygonId < NPolygons);

  loopId = Polygon_nLoops[polygonId];

  assert(Polygon_nLoops[polygonId] < MAX_POLYGON_LOOPS);
  Polygon_nLoops[polygonId]++;

  return loopId;
}

void AddNewVertexInPolygon 
(int polygonId, int loopId,
 double x, double y)
{
  int vertexId;
  
  assert(0 <= polygonId);
  assert(polygonId < NPolygons);
  assert(0 <= loopId);
  assert(loopId < Polygon_nLoops[polygonId]);
  
  vertexId = Polygon_nVertexs[polygonId][loopId];
  
  assert(Polygon_nVertexs[polygonId][loopId] < MAX_POLYGON_VERTEXS);
  Polygon_nVertexs[polygonId][loopId]++;
  
  Polygon_vertexs[polygonId][loopId][vertexId][0] = x;
  Polygon_vertexs[polygonId][loopId][vertexId][1] = y;
}

void WritePolygonFile (const char *fileName)
{
  FILE *fp;
  int iPolygon;
  int iLoop;
  int iVertex;

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  fprintf (fp, "%d\n", 
	   NPolygons);
  
  for (iPolygon = 0; iPolygon < NPolygons; iPolygon++) {
    int nLoops = Polygon_nLoops[iPolygon];
    
    fprintf (fp, "%d   %d\n", 
	     iPolygon,
	     nLoops);
    
    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs = Polygon_nVertexs[iPolygon][iLoop];

      fprintf (fp, "  %d\n", 
	       nVertexs);

      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	double x = Polygon_vertexs[iPolygon][iLoop][iVertex][0];
	double y = Polygon_vertexs[iPolygon][iLoop][iVertex][1];
	
	fprintf (fp, "  %22.15e %22.15e\n", 
		 x, y);
      }
    }
  }

  fclose (fp);
}



/************************
	 Command
*************************/

#define MAX_LOOPS 1000
#define MAX_LOOP_VERTEXS 1000

int NLoops;
double Loop_nVertexs[MAX_LOOPS];
double Loop_vertexs[MAX_LOOPS][MAX_LOOP_VERTEXS][2];
int Loop_parentLoopId[MAX_LOOPS];

static int AddNewLoop (void)
{
  int loopId = NLoops;
  
  assert(NLoops < MAX_LOOPS);
  NLoops++;
  
  return loopId;
}

static void AddNewVertexInLoop 
(int loopId,
 double x, double y)
{
  int vertexId;

  assert(0 <= loopId);
  assert(loopId < NLoops);
  
  vertexId = Loop_nVertexs[loopId];

  assert(Loop_nVertexs[loopId] < MAX_LOOP_VERTEXS);
  Loop_nVertexs[loopId]++;

  Loop_vertexs[loopId][vertexId][0] = x;
  Loop_vertexs[loopId][vertexId][1] = y;
}

static void FindLoop (int startSegId)
{
  int loopId;
  int currentSegId;
  double currentX, currentY;
  int nextSegId;
  double nextX, nextY;
  int iSeg;
  
  loopId = AddNewLoop ();

  {
    double sx1 = Seg_coord1[startSegId][0];
    double sy1 = Seg_coord1[startSegId][1];
    
    MarkSeg (startSegId);
    AddNewVertexInLoop (loopId, 
			sx1, sy1);
    
    currentX = sx1;
    currentY = sy1;
    currentSegId = startSegId;
  }

  for (;;) {

    nextSegId = -1;
    for (iSeg = 0; iSeg < NSegs; iSeg++) {
      if (iSeg != currentSegId) {
	double x0 = Seg_coord0[iSeg][0];
	double y0 = Seg_coord0[iSeg][1];
	
	double x1 = Seg_coord1[iSeg][0];
	double y1 = Seg_coord1[iSeg][1];

	if (PointsAreSame (x0, y0, 
			   currentX, currentY)) {
	  nextSegId = iSeg;
	  nextX = x1;
	  nextY = y1;
	  break;
	}
      }
    }
    assert(0 <= nextSegId);

    if (nextSegId == startSegId) {
      break;
    }

    MarkSeg (nextSegId);
    AddNewVertexInLoop (loopId, 
			nextX, nextY);

    currentSegId = nextSegId;
    currentX = nextX;
    currentY = nextY;
  }
}

static void MakeLoop (void)
{
  for (;;) {
    int segId = GetUnmarkedSeg ();

    if (segId == -1) {
      break;
    }

    FindLoop (segId);
  }
}

static void RemoveFlatAngleVertex (void)
{
  int iLoop;
  int iVertex;

  for (iLoop = 0; iLoop < NLoops; iLoop++) {

    for (;;) {
      int nVertexs = Loop_nVertexs[iLoop];
      int removedVertexId;

      removedVertexId = -1;
      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	int pv = (iVertex - 1 + nVertexs) % nVertexs;
	int cv = iVertex;
	int nv = (iVertex + 1) % nVertexs;
	
	double px = Loop_vertexs[iLoop][pv][0];
	double py = Loop_vertexs[iLoop][pv][1];
	
	double cx = Loop_vertexs[iLoop][cv][0];
	double cy = Loop_vertexs[iLoop][cv][1];
	
	double nx = Loop_vertexs[iLoop][nv][0];
	double ny = Loop_vertexs[iLoop][nv][1];
	
	if (SegContainsPoint (px, py, 
			      nx, ny,
			      cx, cy)) {
	  removedVertexId = iVertex;
	  break;
	}
      }

      if (removedVertexId == -1) {
	break;
      }
      assert(0 <= removedVertexId);

      for (iVertex = removedVertexId; 
	   iVertex < nVertexs -1; 
	   iVertex++) {
	Loop_vertexs[iLoop][iVertex][0] 
	  = Loop_vertexs[iLoop][iVertex + 1][0];
	Loop_vertexs[iLoop][iVertex][1] 
	  = Loop_vertexs[iLoop][iVertex + 1][1];
      }
      Loop_nVertexs[iLoop]--;
    }
  }
}

static void MakePolygon (void)
{
  int iLoop;
  int iVertex;
  int jLoop;

  for (iLoop = 0; iLoop < NLoops; iLoop++) {
    Loop_parentLoopId[iLoop] = -1;
  }

  for (iLoop = 0; iLoop < NLoops; iLoop++) {
    int nVertexs = Loop_nVertexs[iLoop];

    InOut_clear ();
    for (iVertex = 0; iVertex < nVertexs; iVertex++) {
      int v0 = iVertex;
      int v1 = (iVertex + 1) % nVertexs;
      
      double x0 = Loop_vertexs[iLoop][v0][0];
      double y0 = Loop_vertexs[iLoop][v0][1];
      
      double x1 = Loop_vertexs[iLoop][v1][0];
      double y1 = Loop_vertexs[iLoop][v1][1];
      
      InOut_addSeg (x0, y0,
		    x1, y1);
    }

    for (jLoop = 0; jLoop < NLoops; jLoop++) {
      if (jLoop != iLoop) {
	double x = Loop_vertexs[jLoop][0][0];
	double y = Loop_vertexs[jLoop][0][1];

	if (InOut_checkInout (x, y) == INOUT_POINT_INSIDE) {
	  Loop_parentLoopId[jLoop] = iLoop;
	}
      }
    }
  }

  for (iLoop = 0; iLoop < NLoops; iLoop++) {
    int nVertexs = Loop_nVertexs[iLoop];

    if (Loop_parentLoopId[iLoop] == -1) {
      int polygonId = AddNewPolygon ();
      int loopId = AddNewLoopInPolygon (polygonId);

      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	double x = Loop_vertexs[iLoop][iVertex][0];
	double y = Loop_vertexs[iLoop][iVertex][1];
	
	AddNewVertexInPolygon (polygonId, loopId,
			       x, y);
      }

      for (jLoop = 0; jLoop < NLoops; jLoop++) {
	int nVertexs = Loop_nVertexs[jLoop];

	if (Loop_parentLoopId[jLoop] == iLoop) {
	  int loopId = AddNewLoopInPolygon (polygonId);
	  
	  for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	    double x = Loop_vertexs[jLoop][iVertex][0];
	    double y = Loop_vertexs[jLoop][iVertex][1];
	    
	    AddNewVertexInPolygon (polygonId, loopId,
				   x, y);
	  }
	}
      }
    }
  }
}

static void ShiftFirstVertex (void)
{
  int iPolygon;
  int iLoop;
  int iVertex;

  for (iPolygon = 0; iPolygon < NPolygons; iPolygon++) {
    int nLoops = Polygon_nLoops[iPolygon];
    
    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs = Polygon_nVertexs[iPolygon][iLoop];

      for (;;) {
	int v0 = 0;
	int v1 = 1;
	int v2 = nVertexs - 1;
	
	double x0 = Polygon_vertexs[iPolygon][iLoop][v0][0];
	double y0 = Polygon_vertexs[iPolygon][iLoop][v0][1];
	
	double x1 = Polygon_vertexs[iPolygon][iLoop][v1][0];
	double y1 = Polygon_vertexs[iPolygon][iLoop][v1][1];
	
	double x2 = Polygon_vertexs[iPolygon][iLoop][v2][0];
	double y2 = Polygon_vertexs[iPolygon][iLoop][v2][1];

	double nx, ny, nz;

	GetVectorProduct3D
	  (x1 - x0, y1 - y0, 0.0,
	   x2 - x0, y2 - y0, 0.0,
	   &nx, &ny, &nz);
	if (iLoop == 0) {
	  /* outer loop */
	  if (Eps < nz) {
	    break;
	  }
	} else {
	  /* inner loop */
	  if (nz < -Eps) {
	    break;
	  }
	}

	/* rotate vertexs */
	{
	  double x = Polygon_vertexs[iPolygon][iLoop][0][0];
	  double y = Polygon_vertexs[iPolygon][iLoop][0][1];
	  double z = Polygon_vertexs[iPolygon][iLoop][0][2];

	  for (iVertex = 0; iVertex < nVertexs - 1; iVertex++) {
	    Polygon_vertexs[iPolygon][iLoop][iVertex][0]
	      = Polygon_vertexs[iPolygon][iLoop][iVertex + 1][0];
	    Polygon_vertexs[iPolygon][iLoop][iVertex][1]
	      = Polygon_vertexs[iPolygon][iLoop][iVertex + 1][1];
	    Polygon_vertexs[iPolygon][iLoop][iVertex][2]
	      = Polygon_vertexs[iPolygon][iLoop][iVertex + 1][2];
	  }

	  Polygon_vertexs[iPolygon][iLoop][nVertexs - 1][0] = x;
	  Polygon_vertexs[iPolygon][iLoop][nVertexs - 1][1] = y;
	  Polygon_vertexs[iPolygon][iLoop][nVertexs - 1][2] = z;
	}
      }
    }
  }
}

void PerformCommand (void)
{
  MakeLoop ();
  RemoveFlatAngleVertex ();
  MakePolygon ();
  ShiftFirstVertex ();
}



int main (int argc, char *argv[])
{
  if (argc != 3) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautoage_h_gm2d_seg2polygon  seg_file  \n"
       "        polygon_file_OUT \n");
    exit (1);
  }

  ReadSegFile (argv[1]);

  PerformCommand ();

  WritePolygonFile (argv[2]);

  return 0;
}
