/* advautoage_h_gm2d_seg_inout.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

#define DEFAULT_EPS 0.00001

double Eps = DEFAULT_EPS;



void GetLineEquation 
(double x0, double y0, 
 double x1, double y1, 
 double *a_OUT, double *b_OUT, double *c_OUT) 
{
  double dx = x1 - x0;
  double dy = y1 - y0;

  double a0 = -dy;
  double b0 = dx;
  double c0 = dy * x0 - y0 * dx;

  double len = sqrt (a0 * a0 + b0 * b0);

  *a_OUT = a0 / len;  
  *b_OUT = b0 / len;  
  *c_OUT = c0 / len;
}

int Equal (double value0, double value1) 
{
  return fabs (value1 - value0) < Eps;
}

int Between (double value, double range0, double range1) 
{
  return (range0 - Eps < value && value < range1 + Eps)
    || (range1 - Eps < value && value < range0 + Eps);
}

int SegContainsPoint
(double x0, double y0,
 double x1, double y1,
 double x, double y)
{
  double a, b, c;
  
  GetLineEquation 
    (x0, y0,
     x1, y1,
     &a, &b, &c);
      
  if (fabs (a * x + b * y + c) < Eps) {
    /* point is on line of segment */

    if (Equal (x0, x1)) {
      assert(!Equal (y0, y1));
      
      return Between (y, 
		      y0, y1);
    } else {
      return Between (x, 
		      x0, x1);
    }
  } else {
    return 0;
  }
}



/************************
        Segments
*************************/

/* Segment */

#define MAX_SEGS 100000

int NSegs;
double Seg_coord0[MAX_SEGS][2];
double Seg_coord1[MAX_SEGS][2];



void ReadSegFile (const char *fileName)
{
  FILE *fp;
  int iSeg;

  assert(NSegs == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NSegs);
  assert(0 <= NSegs);
  assert(NSegs <= MAX_SEGS);
  
  for (iSeg = 0; iSeg < NSegs; iSeg++) {
    int segId;
    double x0, y0, x1, y1;

    fscanf (fp, "%d %lf %lf %lf %lf", 
	    &segId,
	    &x0, &y0, 
	    &x1, &y1);
    assert(segId == iSeg);

    Seg_coord0[iSeg][0] = x0;
    Seg_coord0[iSeg][1] = y0; 

    Seg_coord1[iSeg][0] = x1;
    Seg_coord1[iSeg][1] = y1; 
  }

  fclose (fp);
}

double Seg_containsPointOnXAxis 
(int segId, 
 double px, double py)
{
  double sx0 = Seg_coord0[segId][0];
  double sx1 = Seg_coord1[segId][0];
  
  return Between (px, 
		  sx0, sx1);
}

double Seg_containsPoint
(int segId, 
 double px, double py)
{
  double sx0 = Seg_coord0[segId][0];
  double sy0 = Seg_coord0[segId][1];
  
  double sx1 = Seg_coord1[segId][0];
  double sy1 = Seg_coord1[segId][1];
  
  return SegContainsPoint
    (sx0, sy0,
     sx1, sy1,
     px, py);
}

#define RAY_UNKNOWN      0
#define RAY_INTERSECT    1
#define RAY_ON_SEGMENT  2

static int CheckSegAndPointAlongYRay 
(int segId,
 double px, double py,
 double *intersectionY_OUT)
{
  int inside2D = Seg_containsPoint (segId,
				    px, py);
  
  if (inside2D) {
    return RAY_ON_SEGMENT;

  } else {
    double sx0 = Seg_coord0[segId][0];
    double sy0 = Seg_coord0[segId][1];
    
    double sx1 = Seg_coord1[segId][0];
    double sy1 = Seg_coord1[segId][1];

    double a, b, c;

    GetLineEquation 
      (sx0, sy0,
       sx1, sy1,
       &a, &b, &c);
    
    if (fabs (b) < Eps) {
      /* parallel to Y axis */
      
      return RAY_UNKNOWN;
    } else {
      int inside2D = Seg_containsPointOnXAxis (segId,
					       px, py);
      double signedDist = a * px + b * py + c;
      
      if (fabs (signedDist) < Eps) {
	/* point is on line of segment */
	/* on boundary, or unknown */
	
	if (inside2D) {
	  return RAY_ON_SEGMENT;
	} else {
	  return RAY_UNKNOWN;
	}
	
      } else {
	/* point is out of line of segment */
	/* intersect or unknown */
	
	if (inside2D) {
	  *intersectionY_OUT = - (c + a * px) / b;

	  if (py < *intersectionY_OUT) {
	    return RAY_INTERSECT;
	  } else {
	    assert(*intersectionY_OUT < py);

	    return RAY_UNKNOWN;
	  }
	} else {
	  return RAY_UNKNOWN;
	}
	
      }
    }
  }

  assert(0);

  return RAY_UNKNOWN;
}

#define MAX_INTERSECTIONS 1000
int NIntersections;
double IntersectionYs[MAX_INTERSECTIONS];

static void ClearIntersection (void)
{
  NIntersections = 0;
}

static void AddIntersection (double y)
{
  int iIntersection;
  int found;
  
  found = 0;
  for (iIntersection = 0; 
       iIntersection < NIntersections; 
       iIntersection++) {
    if (fabs (IntersectionYs[iIntersection] - y) < Eps) {
      found = 1;
      break;
    }
  }
  
  if (!found) {
    assert(NIntersections < MAX_INTERSECTIONS);
    
    IntersectionYs[NIntersections] = y;
    NIntersections++;
  }
}

#define POINT_INOUT_OUTSIDE      0
#define POINT_INOUT_INSIDE       1
#define POINT_INOUT_ON_BOUNDARY  2

int Seg_checkInout (double px, double py)
{  
  int iSeg;

  ClearIntersection ();
  for (iSeg = 0; iSeg < NSegs; iSeg++) {
    double intersectionY;
    int checkResult = CheckSegAndPointAlongYRay 
      (iSeg, 
       px, py,
       &intersectionY);

    switch (checkResult) {
    case RAY_UNKNOWN:
      break;
    case RAY_INTERSECT:
      AddIntersection (intersectionY);
      break;
    case RAY_ON_SEGMENT:
      return POINT_INOUT_ON_BOUNDARY;
    default:
      assert(0);
      break;
    }
  }

  if (NIntersections % 2 == 1) {
    return POINT_INOUT_INSIDE;
  } else {
    return POINT_INOUT_OUTSIDE;
  }
}



/************************
        Points
*************************/

/* Point */

#define MAX_POINTS 1000000


int NPoints;
double Point_coord[MAX_POINTS][2];
int Point_inoutType[MAX_POINTS];



void ReadPointFile (const char *fileName)
{
  FILE *fp;
  int iPoint;
  
  assert(NPoints == 0);
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NPoints);
  assert(0 <= NPoints);
  assert(NPoints <= MAX_POINTS);
  
  for (iPoint = 0; iPoint < NPoints; iPoint++) {
    int pointId;
    double x, y;

    fscanf (fp, "%d %lf %lf", 
	    &pointId,
	    &x, &y);
    assert(pointId == iPoint);
    Point_coord[iPoint][0] = x;
    Point_coord[iPoint][1] = y; 
  }

  fclose (fp);
}

void WritePointInoutFile (const char *fileName)
{
  FILE *fp;
  int iPoint;
  
  assert(0 < NPoints);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NPoints);
  
  for (iPoint = 0; iPoint < NPoints; iPoint++) {
    fprintf (fp, "%d   %d\n", 
	     iPoint, Point_inoutType[iPoint]);
  }

  fclose (fp);
}



/************************
	 Command
*************************/

void PerformCommand (void)
{
  int iPoint;
  
  for (iPoint = 0; iPoint < NPoints; iPoint++) {
    double px = Point_coord[iPoint][0];
    double py = Point_coord[iPoint][1];
    
    int inoutType0 = Seg_checkInout (px, py);

    if (inoutType0 == POINT_INOUT_ON_BOUNDARY) {
      Point_inoutType[iPoint] = POINT_INOUT_ON_BOUNDARY;

    } else {
      double delta = Eps * 10;
      int inoutType1 = Seg_checkInout (px - delta, py);

      if (inoutType1 == inoutType0) {
	Point_inoutType[iPoint] = inoutType0;
      } else {
	int inoutType2 = Seg_checkInout (px + delta, py);
	int insideCount;
	
	insideCount = 0;
	if (inoutType0 == POINT_INOUT_INSIDE) {
	  insideCount++;
	}
	if (inoutType1 == POINT_INOUT_INSIDE) {
	  insideCount++;
	}
	if (inoutType2 == POINT_INOUT_INSIDE) {
	  insideCount++;
	}
	
	if (2 <= insideCount) {
	  Point_inoutType[iPoint] = POINT_INOUT_INSIDE;
	} else {
	  Point_inoutType[iPoint] = POINT_INOUT_OUTSIDE;
	}
      }
    }
  }
}



int main (int argc, char *argv[])
{
  if (argc != 4) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautoage_h_gm2d_seg_inout  seg_file  \n"
       "        point_file \n"
       "        point_inout_file_OUT \n");
    exit (1);
  }

  ReadSegFile (argv[1]);
  ReadPointFile (argv[2]);

  PerformCommand ();

  WritePointInoutFile (argv[3]);

  return 0;
}
