/* advautoage_h_gm2d_set_polygon.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

#define DEFAULT_EPS 0.00001

double Eps = DEFAULT_EPS;



int Equal (double value0, double value1) 
{
  return fabs (value1 - value0) < Eps;
}

int Between (double value, double range0, double range1) 
{
  return (range0 - Eps < value && value < range1 + Eps)
    || (range1 - Eps < value && value < range0 + Eps);
}

double Distance (double x0, double y0, 
		 double x1, double y1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) + (y1 - y0) * (y1 - y0));
}

int PointsAreSame (double x0, double y0, 
		   double x1, double y1) 
{
  return Distance (x0, y0, 
		   x1, y1) < Eps;
}

int SegsAreSame 
(double seg0_x0, double seg0_y0, 
 double seg0_x1, double seg0_y1, 
 double seg1_x0, double seg1_y0, 
 double seg1_x1, double seg1_y1) 
{
  if (PointsAreSame (seg0_x0, seg0_y0,
		     seg1_x0, seg1_y0)
      && PointsAreSame (seg0_x1, seg0_y1,
			seg1_x1, seg1_y1)) {
    return 1;
  }
  if (PointsAreSame (seg0_x0, seg0_y0,
		     seg1_x1, seg1_y1)
      && PointsAreSame (seg0_x1, seg0_y1,
			seg1_x0, seg1_y0)) {
    return 1;
  }
  return 0;
}

void GetLineEquation 
(double x0, double y0, 
 double x1, double y1, 
 double *a_OUT, double *b_OUT, double *c_OUT) 
{
  double dx = x1 - x0;
  double dy = y1 - y0;

  double a0 = -dy;
  double b0 = dx;
  double c0 = dy * x0 - y0 * dx;

  double len = sqrt (a0 * a0 + b0 * b0);

  *a_OUT = a0 / len;  
  *b_OUT = b0 / len;  
  *c_OUT = c0 / len;
}

int SegContainsPoint
(double x0, double y0,
 double x1, double y1,
 double x, double y)
{
  double a, b, c;
  
  GetLineEquation 
    (x0, y0,
     x1, y1,
     &a, &b, &c);
      
  if (fabs (a * x + b * y + c) < Eps) {
    /* point is on line of segment */

    if (Equal (x0, x1)) {
      assert(!Equal (y0, y1));
      
      return Between (y, 
		      y0, y1);
    } else {
      return Between (x, 
		      x0, x1);
    }
  } else {
    return 0;
  }
}



/************************
     In-Out Segments
*************************/

#define MAX_INOUT_SEGS 100000

static int InOut_nSegs;
static double InOut_seg_coord0[MAX_INOUT_SEGS][2];
static double InOut_seg_coord1[MAX_INOUT_SEGS][2];

#define MAX_INOUT_INTERSECTIONS 1000
static int InOut_nIntersections;
static double InOut_intersectionYs[MAX_INOUT_INTERSECTIONS];



static double InOut_containsPointOnXAxis 
(int segId, 
 double px, double py)
{
  double sx0 = InOut_seg_coord0[segId][0];
  double sx1 = InOut_seg_coord1[segId][0];
  
  return Between (px, 
		  sx0, sx1);
}

static double InOut_containsPoint
(int segId, 
 double px, double py)
{
  double sx0 = InOut_seg_coord0[segId][0];
  double sy0 = InOut_seg_coord0[segId][1];
  
  double sx1 = InOut_seg_coord1[segId][0];
  double sy1 = InOut_seg_coord1[segId][1];
  
  return SegContainsPoint
    (sx0, sy0,
     sx1, sy1,
     px, py);
}

#define INOUT_RAY_UNKNOWN      0
#define INOUT_RAY_INTERSECT    1
#define INOUT_RAY_ON_SEGMENT  2

static int InOut_checkSegAndPointAlongYRay 
(int segId,
 double px, double py,
 double *intersectionY_OUT)
{
  int inside2D = InOut_containsPoint (segId,
				      px, py);
  
  if (inside2D) {
    return INOUT_RAY_ON_SEGMENT;

  } else {
    double sx0 = InOut_seg_coord0[segId][0];
    double sy0 = InOut_seg_coord0[segId][1];
    
    double sx1 = InOut_seg_coord1[segId][0];
    double sy1 = InOut_seg_coord1[segId][1];

    double a, b, c;

    GetLineEquation 
      (sx0, sy0,
       sx1, sy1,
       &a, &b, &c);
    
    if (fabs (b) < Eps) {
      /* parallel to Y axis */
      
      return INOUT_RAY_UNKNOWN;
    } else {
      int inside2D = InOut_containsPointOnXAxis (segId,
						 px, py);
      double signedDist = a * px + b * py + c;
      
      if (fabs (signedDist) < Eps) {
	/* point is on line of segment */
	/* on boundary, or unknown */
	
	if (inside2D) {
	  return INOUT_RAY_ON_SEGMENT;
	} else {
	  return INOUT_RAY_UNKNOWN;
	}
	
      } else {
	/* point is out of line of segment */
	/* intersect or unknown */
	
	if (inside2D) {
	  *intersectionY_OUT = - (c + a * px) / b;

	  if (py < *intersectionY_OUT) {
	    return INOUT_RAY_INTERSECT;
	  } else {
	    assert(*intersectionY_OUT < py);

	    return INOUT_RAY_UNKNOWN;
	  }
	} else {
	  return INOUT_RAY_UNKNOWN;
	}
	
      }
    }
  }

  assert(0);

  return INOUT_RAY_UNKNOWN;
}

static void InOut_clearIntersection (void)
{
  InOut_nIntersections = 0;
}

static void InOut_addIntersection (double y)
{
  int iIntersection;
  int found;
  
  found = 0;
  for (iIntersection = 0; 
       iIntersection < InOut_nIntersections; 
       iIntersection++) {
    if (fabs (InOut_intersectionYs[iIntersection] - y) < Eps) {
      found = 1;
      break;
    }
  }
  
  if (!found) {
    assert(InOut_nIntersections < MAX_INOUT_INTERSECTIONS);
    
    InOut_intersectionYs[InOut_nIntersections] = y;
    InOut_nIntersections++;
  }
}

void InOut_clear (void)
{
  InOut_nSegs = 0;
}

void InOut_addSeg 
(double x0, double y0,
 double x1, double y1)
{
  assert(InOut_nSegs < MAX_INOUT_SEGS);

  InOut_seg_coord0[InOut_nSegs][0] = x0;
  InOut_seg_coord0[InOut_nSegs][1] = y0;
  
  InOut_seg_coord1[InOut_nSegs][0] = x1;
  InOut_seg_coord1[InOut_nSegs][1] = y1;

  InOut_nSegs++;
}

#define INOUT_POINT_OUTSIDE      0
#define INOUT_POINT_INSIDE       1
#define INOUT_POINT_ON_BOUNDARY  2

int InOut_checkInout (double px, double py)
{  
  int iSeg;

  InOut_clearIntersection ();
  for (iSeg = 0; iSeg < InOut_nSegs; iSeg++) {
    double intersectionY;
    int checkResult = InOut_checkSegAndPointAlongYRay 
      (iSeg, 
       px, py,
       &intersectionY);

    switch (checkResult) {
    case INOUT_RAY_UNKNOWN:
      break;
    case INOUT_RAY_INTERSECT:
      InOut_addIntersection (intersectionY);
      break;
    case INOUT_RAY_ON_SEGMENT:
      return INOUT_POINT_ON_BOUNDARY;
    default:
      assert(0);
      break;
    }
  }

  if (InOut_nIntersections % 2 == 1) {
    return INOUT_POINT_INSIDE;
  } else {
    return INOUT_POINT_OUTSIDE;
  }
}



/************************
        Segments
*************************/

/* Segment */

#define MAX_SEGS 100000

int NSegsA;
double SegA_coord0[MAX_SEGS][2];
double SegA_coord1[MAX_SEGS][2];
int SegA_insideIsInB[MAX_SEGS];
int SegA_outsideIsInB[MAX_SEGS];
int SegA_isInResult[MAX_SEGS];

int NSegsB;
double SegB_coord0[MAX_SEGS][2];
double SegB_coord1[MAX_SEGS][2];
int SegB_insideIsInA[MAX_SEGS];
int SegB_outsideIsInA[MAX_SEGS];
int SegB_isInResult[MAX_SEGS];



void ReadSegFileA (const char *fileName)
{
  FILE *fp;
  int iSeg;

  assert(NSegsA == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NSegsA);
  assert(0 <= NSegsA);
  assert(NSegsA <= MAX_SEGS);
  
  for (iSeg = 0; iSeg < NSegsA; iSeg++) {
    int segId;
    double x0, y0, x1, y1;

    fscanf (fp, "%d %lf %lf %lf %lf", 
	    &segId,
	    &x0, &y0, 
	    &x1, &y1);
    assert(segId == iSeg);

    SegA_coord0[iSeg][0] = x0;
    SegA_coord0[iSeg][1] = y0; 

    SegA_coord1[iSeg][0] = x1;
    SegA_coord1[iSeg][1] = y1; 
  }

  fclose (fp);
}

void ReadSegFileB (const char *fileName)
{
  FILE *fp;
  int iSeg;

  assert(NSegsB == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NSegsB);
  assert(0 <= NSegsB);
  assert(NSegsB <= MAX_SEGS);
  
  for (iSeg = 0; iSeg < NSegsB; iSeg++) {
    int segId;
    double x0, y0, x1, y1;

    fscanf (fp, "%d %lf %lf %lf %lf", 
	    &segId,
	    &x0, &y0, 
	    &x1, &y1);
    assert(segId == iSeg);

    SegB_coord0[iSeg][0] = x0;
    SegB_coord0[iSeg][1] = y0; 

    SegB_coord1[iSeg][0] = x1;
    SegB_coord1[iSeg][1] = y1; 
  }

  fclose (fp);
}



/************************
	 Polygons
*************************/

#define MAX_POLYGONS 1000
#define MAX_POLYGON_LOOPS 100
#define MAX_POLYGON_VERTEXS 100

int NPolygonsA;
int PolygonA_nLoops[MAX_POLYGONS];
int PolygonA_nVertexs[MAX_POLYGONS][MAX_POLYGON_LOOPS];
double PolygonA_vertexs[MAX_POLYGONS][MAX_POLYGON_LOOPS][MAX_POLYGON_VERTEXS][2];

int NPolygonsB;
int PolygonB_nLoops[MAX_POLYGONS];
int PolygonB_nVertexs[MAX_POLYGONS][MAX_POLYGON_LOOPS];
double PolygonB_vertexs[MAX_POLYGONS][MAX_POLYGON_LOOPS][MAX_POLYGON_VERTEXS][2];



void ReadPolygonFileA (const char *fileName)
{
  FILE *fp;
  int iPolygon;
  int iLoop;
  int iVertex;

  assert(NPolygonsA == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NPolygonsA);
  assert(0 <= NPolygonsA);
  assert(NPolygonsA <= MAX_POLYGONS);
  
  for (iPolygon = 0; iPolygon < NPolygonsA; iPolygon++) {
    int polygonId;
    int nLoops;
    
    fscanf (fp, "%d %d", 
	    &polygonId,
	    &nLoops);
    assert(polygonId == iPolygon);
    assert(0 < nLoops);
    
    PolygonA_nLoops[iPolygon] = nLoops;
    
    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs;

      fscanf (fp, "%d", 
	      &nVertexs);
      assert(2 < nVertexs);
      
      PolygonA_nVertexs[iPolygon][iLoop] = nVertexs;

      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	double x, y;

	fscanf (fp, "%lf %lf", 
		&x, &y);
	
	PolygonA_vertexs[iPolygon][iLoop][iVertex][0] = x;
	PolygonA_vertexs[iPolygon][iLoop][iVertex][1] = y; 
      }
    }
  }

  fclose (fp);
}

void ReadPolygonFileB (const char *fileName)
{
  FILE *fp;
  int iPolygon;
  int iLoop;
  int iVertex;

  assert(NPolygonsB == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NPolygonsB);
  assert(0 <= NPolygonsB);
  assert(NPolygonsB <= MAX_POLYGONS);
  
  for (iPolygon = 0; iPolygon < NPolygonsB; iPolygon++) {
    int polygonId;
    int nLoops;
    
    fscanf (fp, "%d %d", 
	    &polygonId,
	    &nLoops);
    assert(polygonId == iPolygon);
    assert(0 < nLoops);
    
    PolygonB_nLoops[iPolygon] = nLoops;
    
    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs;

      fscanf (fp, "%d", 
	      &nVertexs);
      assert(2 < nVertexs);
      
      PolygonB_nVertexs[iPolygon][iLoop] = nVertexs;

      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	double x, y;

	fscanf (fp, "%lf %lf", 
		&x, &y);
	
	PolygonB_vertexs[iPolygon][iLoop][iVertex][0] = x;
	PolygonB_vertexs[iPolygon][iLoop][iVertex][1] = y; 
      }
    }
  }

  fclose (fp);
}

void WritePolygonSegFileA (const char *fileName)
{
  FILE *fp;
  int iPolygon;
  int iLoop;
  int iVertex;
  int count;

  assert(0 <= NPolygonsA);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  count = 0;
  for (iPolygon = 0; iPolygon < NPolygonsA; iPolygon++) {
    int nLoops = PolygonA_nLoops[iPolygon];
    
    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs = PolygonA_nVertexs[iPolygon][iLoop];
      
      count += nVertexs;
    }
  }

  fprintf (fp, "%d\n", 
	   count);

  count = 0;
  for (iPolygon = 0; iPolygon < NPolygonsA; iPolygon++) {
    int nLoops = PolygonA_nLoops[iPolygon];
    
    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs = PolygonA_nVertexs[iPolygon][iLoop];
    
      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	int v0 = iVertex;
	int v1 = (iVertex + 1) % nVertexs;
	
	double x0 = PolygonA_vertexs[iPolygon][iLoop][v0][0];
	double y0 = PolygonA_vertexs[iPolygon][iLoop][v0][1];
	
	double x1 = PolygonA_vertexs[iPolygon][iLoop][v1][0];
	double y1 = PolygonA_vertexs[iPolygon][iLoop][v1][1];
	
	fprintf (fp, "%d   %22.15e %22.15e   %22.15e %22.15e\n", 
		 count,
		 x0, y0,
		 x1, y1);
	count++;
      }
    }
  }

  fclose (fp);
}

void WritePolygonSegFileB (const char *fileName)
{
  FILE *fp;
  int iPolygon;
  int iLoop;
  int iVertex;
  int count;

  assert(0 <= NPolygonsB);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  count = 0;
  for (iPolygon = 0; iPolygon < NPolygonsB; iPolygon++) {
    int nLoops = PolygonB_nLoops[iPolygon];
    
    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs = PolygonB_nVertexs[iPolygon][iLoop];
      
      count += nVertexs;
    }
  }

  fprintf (fp, "%d\n", 
	   count);

  count = 0;
  for (iPolygon = 0; iPolygon < NPolygonsB; iPolygon++) {
    int nLoops = PolygonB_nLoops[iPolygon];
    
    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs = PolygonB_nVertexs[iPolygon][iLoop];
    
      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	int v0 = iVertex;
	int v1 = (iVertex + 1) % nVertexs;
	
	double x0 = PolygonB_vertexs[iPolygon][iLoop][v0][0];
	double y0 = PolygonB_vertexs[iPolygon][iLoop][v0][1];
	
	double x1 = PolygonB_vertexs[iPolygon][iLoop][v1][0];
	double y1 = PolygonB_vertexs[iPolygon][iLoop][v1][1];
	
	fprintf (fp, "%d   %22.15e %22.15e   %22.15e %22.15e\n", 
		 count,
		 x0, y0,
		 x1, y1);
	count++;
      }
    }
  }

  fclose (fp);
}



/************************
	 Command
*************************/

#define POLYGON_SEG_A_FILE_NAME "___set_polygon__ps_a.seg"
#define POLYGON_SEG_B_FILE_NAME "___set_polygon__ps_b.seg"

#define SEG_A_FILE_NAME "___set_polygon__seg_a.seg"
#define SEG_B_FILE_NAME "___set_polygon__seg_b.seg"

#define SEG_AB_FILE_NAME "___set_polygon__seg_ab.seg"

#define DIVIDE_COMMAND "advautoage_h_gm2d_divide_seg_by_seg"
#define SEG2POLYGON_COMMAND "advautoage_h_gm2d_seg2polygon"

static void DivideSeg (void)
{
  char buf[2048];

  WritePolygonSegFileA (POLYGON_SEG_A_FILE_NAME);
  WritePolygonSegFileB (POLYGON_SEG_B_FILE_NAME);

  sprintf (buf, "%s %s %s %s",
	   DIVIDE_COMMAND,
	   POLYGON_SEG_A_FILE_NAME,
	   POLYGON_SEG_B_FILE_NAME,
	   SEG_A_FILE_NAME); 
  system (buf);

  sprintf (buf, "%s %s %s %s",
	   DIVIDE_COMMAND,
	   POLYGON_SEG_B_FILE_NAME,
	   POLYGON_SEG_A_FILE_NAME,
	   SEG_B_FILE_NAME); 
  system (buf);

  ReadSegFileA (SEG_A_FILE_NAME);
  ReadSegFileB (SEG_B_FILE_NAME);
}

static void CheckInOutOfSegA (void)
{
  int iPolygon;
  int iLoop;
  int iVertex;
  int iSeg;
  
  InOut_clear ();
  for (iPolygon = 0; iPolygon < NPolygonsB; iPolygon++) {
    int nLoops = PolygonB_nLoops[iPolygon];
    
    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs = PolygonB_nVertexs[iPolygon][iLoop];
      
      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	int v0 = iVertex;
	int v1 = (iVertex + 1) % nVertexs;
	
	double x0 = PolygonB_vertexs[iPolygon][iLoop][v0][0];
	double y0 = PolygonB_vertexs[iPolygon][iLoop][v0][1];
	
	double x1 = PolygonB_vertexs[iPolygon][iLoop][v1][0];
	double y1 = PolygonB_vertexs[iPolygon][iLoop][v1][1];
	
	InOut_addSeg (x0, y0,
		      x1, y1);
      }
    }
  }

  for (iSeg = 0; iSeg < NSegsA; iSeg++) {
    double x0 = SegA_coord0[iSeg][0];
    double y0 = SegA_coord0[iSeg][1]; 
    
    double x1 = SegA_coord1[iSeg][0];
    double y1 = SegA_coord1[iSeg][1]; 

    double cx = (x0 + x1) / 2;
    double cy = (y0 + y1) / 2;

    double a, b, c;
    double sampleLength;
    double insideX, insideY;
    double outsideX, outsideY;

    GetLineEquation (x0, y0,
		     x1, y1,
		     &a, &b, &c);
    
    /* NEED TO REDEFINE */
    sampleLength = Eps * 100;

    insideX = cx + a * sampleLength;
    insideY = cy + b * sampleLength;
    if (InOut_checkInout (insideX, insideY) == INOUT_POINT_INSIDE) {
      SegA_insideIsInB[iSeg] = 1;
    } else {
      SegA_insideIsInB[iSeg] = 0;
    }

    outsideX = cx - a * sampleLength;
    outsideY = cy - b * sampleLength;
    if (InOut_checkInout (outsideX, outsideY) == INOUT_POINT_INSIDE) {
      SegA_outsideIsInB[iSeg] = 1;
    } else {
      SegA_outsideIsInB[iSeg] = 0;
    }
  }
}

static void CheckInOutOfSegB (void)
{
  int iPolygon;
  int iLoop;
  int iVertex;
  int iSeg;
  
  InOut_clear ();
  for (iPolygon = 0; iPolygon < NPolygonsA; iPolygon++) {
    int nLoops = PolygonA_nLoops[iPolygon];
    
    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs = PolygonA_nVertexs[iPolygon][iLoop];
    
      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	int v0 = iVertex;
	int v1 = (iVertex + 1) % nVertexs;
	
	double x0 = PolygonA_vertexs[iPolygon][iLoop][v0][0];
	double y0 = PolygonA_vertexs[iPolygon][iLoop][v0][1];
	
	double x1 = PolygonA_vertexs[iPolygon][iLoop][v1][0];
	double y1 = PolygonA_vertexs[iPolygon][iLoop][v1][1];
      
	InOut_addSeg (x0, y0,
		      x1, y1);
      }
    }
  }

  for (iSeg = 0; iSeg < NSegsB; iSeg++) {
    double x0 = SegB_coord0[iSeg][0];
    double y0 = SegB_coord0[iSeg][1]; 
    
    double x1 = SegB_coord1[iSeg][0];
    double y1 = SegB_coord1[iSeg][1]; 

    double cx = (x0 + x1) / 2;
    double cy = (y0 + y1) / 2;

    double a, b, c;
    double sampleLength;
    double insideX, insideY;
    double outsideX, outsideY;

    GetLineEquation (x0, y0,
		     x1, y1,
		     &a, &b, &c);
    
    /* NEED TO REDEFINE */
    sampleLength = Eps * 100;

    insideX = cx + a * sampleLength;
    insideY = cy + b * sampleLength;
    if (InOut_checkInout (insideX, insideY) == INOUT_POINT_INSIDE) {
      SegB_insideIsInA[iSeg] = 1;
    } else {
      SegB_insideIsInA[iSeg] = 0;
    }

    outsideX = cx - a * sampleLength;
    outsideY = cy - b * sampleLength;
    if (InOut_checkInout (outsideX, outsideY) == INOUT_POINT_INSIDE) {
      SegB_outsideIsInA[iSeg] = 1;
    } else {
      SegB_outsideIsInA[iSeg] = 0;
    }
  }
}

static void CollectSegForUnion (void)
{
  int iSeg;

  for (iSeg = 0; iSeg < NSegsA; iSeg++) {
    SegA_isInResult[iSeg] 
      = !SegA_outsideIsInB[iSeg];
  }
  for (iSeg = 0; iSeg < NSegsB; iSeg++) {
    SegB_isInResult[iSeg] 
      = !SegB_outsideIsInA[iSeg];
  }
}

static void CollectSegForIntersection (void)
{
  int iSeg;

  for (iSeg = 0; iSeg < NSegsA; iSeg++) {
    if (SegA_insideIsInB[iSeg]
	&& !SegA_outsideIsInB[iSeg]) {
      /* seg A is on boundary of polygon B */

      SegA_isInResult[iSeg] = 1;
    } else {
      SegA_isInResult[iSeg] 
	= SegA_outsideIsInB[iSeg];
    }
  }

  for (iSeg = 0; iSeg < NSegsB; iSeg++) {
    if (SegB_insideIsInA[iSeg]
	&& !SegB_outsideIsInA[iSeg]) {
      /* seg B is on boundary of polygon A */

      SegB_isInResult[iSeg] = 1;
    } else {
      SegB_isInResult[iSeg] 
	= SegB_outsideIsInA[iSeg];
    }
  }
}

static void CollectSegForSubtraction (void)
{
  int iSeg;

  for (iSeg = 0; iSeg < NSegsA; iSeg++) {
    if (SegA_insideIsInB[iSeg]
	&& !SegA_outsideIsInB[iSeg]) {
      /* seg A is on boundary of polygon B */

      SegA_isInResult[iSeg] = 0;
    } else {
      SegA_isInResult[iSeg] 
	= !SegA_outsideIsInB[iSeg];
    }
  }

  for (iSeg = 0; iSeg < NSegsB; iSeg++) {
    double dummyX = SegB_coord0[iSeg][0];
    double dummyY = SegB_coord0[iSeg][1];
    
    SegB_coord0[iSeg][0] = SegB_coord1[iSeg][0];
    SegB_coord0[iSeg][1] = SegB_coord1[iSeg][1];
    
    SegB_coord1[iSeg][0] = dummyX;
    SegB_coord1[iSeg][1] = dummyY;
    
    SegB_isInResult[iSeg] 
      = SegB_outsideIsInA[iSeg];
  }
}

static void RemoveDuplicatedSeg (void)
{
  int iSeg;
  int jSeg;
  int found;

  for (iSeg = 0; iSeg < NSegsA; iSeg++) {
    if (SegA_isInResult[iSeg]) {
      double ax0 = SegA_coord0[iSeg][0];
      double ay0 = SegA_coord0[iSeg][1]; 
      
      double ax1 = SegA_coord1[iSeg][0];
      double ay1 = SegA_coord1[iSeg][1]; 

      found = 0;
      for (jSeg = 0; jSeg < NSegsB; jSeg++) {
	if (SegB_isInResult[jSeg]) {
	  double bx0 = SegB_coord0[jSeg][0];
	  double by0 = SegB_coord0[jSeg][1]; 
	  
	  double bx1 = SegB_coord1[jSeg][0];
	  double by1 = SegB_coord1[jSeg][1]; 

	  if (SegsAreSame (bx0, by0,
			   bx1, by1,
			   ax0, ay0,
			   ax1, ay1)) {
	    found = 1;
	    break;
	  }
	}
      }
      if (found) {
	SegA_isInResult[iSeg] = 0;
      }
    }
  }
}

static void WriteCollectedSeg (void)
{
  FILE *fp;
  int iSeg;
  int count;

  fp = fopen (SEG_AB_FILE_NAME, "w");
  assert(fp != NULL);
  
  count = 0;
  for (iSeg = 0; iSeg < NSegsA; iSeg++) {
    if (SegA_isInResult[iSeg]) {
      count++;
    }
  }
  for (iSeg = 0; iSeg < NSegsB; iSeg++) {
    if (SegB_isInResult[iSeg]) {
      count++;
    }
  }

  fprintf (fp, "%d\n", 
	   count);

  count = 0;
  for (iSeg = 0; iSeg < NSegsA; iSeg++) {
    if (SegA_isInResult[iSeg]) {
      double x0 = SegA_coord0[iSeg][0];
      double y0 = SegA_coord0[iSeg][1];

      double x1 = SegA_coord1[iSeg][0];
      double y1 = SegA_coord1[iSeg][1];

      fprintf (fp, "%d   %22.15e %22.15e   %22.15e %22.15e\n", 
	       count,
	       x0, y0,
	       x1, y1);
      count++;
    }
  }
  for (iSeg = 0; iSeg < NSegsB; iSeg++) {
    if (SegB_isInResult[iSeg]) {
      double x0 = SegB_coord0[iSeg][0];
      double y0 = SegB_coord0[iSeg][1];

      double x1 = SegB_coord1[iSeg][0];
      double y1 = SegB_coord1[iSeg][1];

      fprintf (fp, "%d   %22.15e %22.15e   %22.15e %22.15e\n", 
	       count,
	       x0, y0,
	       x1, y1);
      count++;
    }
  }

  fclose (fp);
}

static void MakePolygon (const char *outputPolygonFileName)
{
  char buf[2048];

  sprintf (buf, "%s %s %s",
	   SEG2POLYGON_COMMAND,
	   SEG_AB_FILE_NAME,
	   outputPolygonFileName); 
  system (buf);
}

void PerformCommand 
(const char *setOperationType,
 const char *outputPolygonFileName)
{
  DivideSeg ();
  CheckInOutOfSegA ();
  CheckInOutOfSegB ();
  
  if (strcmp (setOperationType, "add") == 0) {
    CollectSegForUnion ();
  } else if (strcmp (setOperationType, "mul") == 0) {
    CollectSegForIntersection ();
  } else if (strcmp (setOperationType, "sub") == 0) {
    CollectSegForSubtraction ();
  } else {
    assert(0);
  }
  
  RemoveDuplicatedSeg ();
  WriteCollectedSeg ();
  MakePolygon (outputPolygonFileName);
}



int main (int argc, char *argv[])
{
  if (argc != 5) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautoage_h_gm2d_set_polygon  polygon_file_a  polygon_file_b \n"
       "        set_operation_type \n"
       "        polygon_file_OUT  \n"
       "                          \n"
       "    set_operation_type :  \n"
       "      add   (union)         \n"
       "      mul   (intersection)  \n"
       "      sub   (subtraction)   \n");
    exit (1);
  }

  ReadPolygonFileA (argv[1]);
  ReadPolygonFileB (argv[2]);

  PerformCommand (argv[3], argv[4]);

  return 0;
}
