/* advautoage_h_gm2d_tessellate.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

#define DEFAULT_EPS 0.00001

double Eps = DEFAULT_EPS;



double Distance (double x0, double y0, 
		 double x1, double y1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) + (y1 - y0) * (y1 - y0));
}

int PointsAreSame (double x0, double y0, 
		   double x1, double y1) 
{
  return Distance (x0, y0, 
		   x1, y1) < Eps;
}

double SignedTriArea
(double x0, double y0,
 double x1, double y1,
 double x2, double y2)
{
  return ((x1 - x0) * (y2 - y0) - (y1 - y0) * (x2 - x0)) / 2;
}

void GetLineEquation 
(double x0, double y0, 
 double x1, double y1, 
 double *a_OUT, double *b_OUT, double *c_OUT) 
{
  double dx = x1 - x0;
  double dy = y1 - y0;

  double a0 = -dy;
  double b0 = dx;
  double c0 = dy * x0 - y0 * dx;

  double len = sqrt (a0 * a0 + b0 * b0);

  *a_OUT = a0 / len;  
  *b_OUT = b0 / len;  
  *c_OUT = c0 / len;
}

int SolveLinearEquation
(double a00, double a01,
 double a10, double a11, 
 double b0, double b1,
 double *x_OUT, double *y_OUT)
{
  double det = a00 * a11 - a01 * a10;

  /* a00 x + a01 y + b0 = 0 */
  /* a10 x + a11 y + b1 = 0 */

  if (fabs (det) < Eps * Eps) return 0;

  *x_OUT = (a01 * b1 - a11 * b0) / det;
  *y_OUT = (a10 * b0 - a00 * b1) / det;
  return 1;
}

int Between (double value, double range0, double range1) 
{
  return (range0 - Eps < value && value < range1 + Eps)
    || (range1 - Eps < value && value < range0 + Eps);
}

int GetIntersectionBetweenSegs 
(double seg0_x0, double seg0_y0, 
 double seg0_x1, double seg0_y1, 
 double seg1_x0, double seg1_y0, 
 double seg1_x1, double seg1_y1,
 double *intX_OUT, double *intY_OUT) 
{
  double a0, b0, c0;
  double a1, b1, c1;
  int flag;
  
  GetLineEquation (seg0_x0, seg0_y0, 
		   seg0_x1, seg0_y1, 
		   &a0, &b0, &c0);
  GetLineEquation (seg1_x0, seg1_y0, 
		   seg1_x1, seg1_y1, 
		   &a1, &b1, &c1);
  flag = SolveLinearEquation 
    (a0, b0, 
     a1, b1, 
     c0, c1,
     intX_OUT, intY_OUT);
  if (flag) {
    if (Between (*intX_OUT, seg0_x0, seg0_x1)
	&& Between (*intY_OUT, seg0_y0, seg0_y1)) {
      if (Between (*intX_OUT, seg1_x0, seg1_x1)
	  && Between (*intY_OUT, seg1_y0, seg1_y1)) {
	return 1;
      }
    }
  }
  return 0;
}



/************************
        Segments
*************************/

/* Segment */

#define MAX_SEGS 100000

int NSegs;
double Seg_coord0[MAX_SEGS][2];
double Seg_coord1[MAX_SEGS][2];



void ReadSegFile (const char *fileName)
{
  FILE *fp;
  int iSeg;

  assert(NSegs == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NSegs);
  assert(0 <= NSegs);
  assert(NSegs <= MAX_SEGS);
  
  for (iSeg = 0; iSeg < NSegs; iSeg++) {
    int segId;
    double x0, y0, x1, y1;

    fscanf (fp, "%d %lf %lf %lf %lf", 
	    &segId,
	    &x0, &y0, 
	    &x1, &y1);
    assert(segId == iSeg);

    Seg_coord0[iSeg][0] = x0;
    Seg_coord0[iSeg][1] = y0; 

    Seg_coord1[iSeg][0] = x1;
    Seg_coord1[iSeg][1] = y1; 
  }

  fclose (fp);
}

int AddNewSeg 
(double x0, double y0,
 double x1, double y1)
{
  int segId = NSegs;

  assert(NSegs < MAX_SEGS);
  NSegs++;

  Seg_coord0[segId][0] = x0;
  Seg_coord0[segId][1] = y0;

  Seg_coord1[segId][0] = x1;
  Seg_coord1[segId][1] = y1;

  return segId;
}

int GetSeg (void)
{
  if (0 < NSegs) {
    return 0;
  } else {
    return -1;
  }
}

int FindSeg (double sx0, double sy0, 
	     double sx1, double sy1)
{
  int iSeg;
  
  for (iSeg = 0; iSeg < NSegs; iSeg++) {
    double x0 = Seg_coord0[iSeg][0];
    double y0 = Seg_coord0[iSeg][1]; 
    
    double x1 = Seg_coord1[iSeg][0];
    double y1 = Seg_coord1[iSeg][1]; 

    if ((PointsAreSame (x0, y0, 
			sx0, sy0) 
	 && PointsAreSame (x1, y1, 
			   sx1, sy1)) 
	|| (PointsAreSame (x0, y0, 
			   sx1, sy1) 
	    && PointsAreSame (x1, y1, 
			      sx0, sy0))) {
      return iSeg;
    }
  }
  return -1;
}

void RemoveSeg 
(double x0, double y0,
 double x1, double y1)
{
  int segId = FindSeg (x0, y0,
		       x1, y1);
  int iSeg;
  
  assert(0 <= segId);
  assert(segId < NSegs);
  
  for (iSeg = segId; iSeg < NSegs - 1; iSeg++) {
    Seg_coord0[iSeg][0] = Seg_coord0[iSeg + 1][0];
    Seg_coord0[iSeg][1] = Seg_coord0[iSeg + 1][1];
    
    Seg_coord1[iSeg][0] = Seg_coord1[iSeg + 1][0];
    Seg_coord1[iSeg][1] = Seg_coord1[iSeg + 1][1];
  }
  NSegs--;
}




/************************
        Triangles
*************************/

/* Triangle */

#define MAX_TRIS 100000

int NTris;
double Tri_coord0[MAX_TRIS][2];
double Tri_coord1[MAX_TRIS][2];
double Tri_coord2[MAX_TRIS][2];



int AddNewTri 
(double x0, double y0,
 double x1, double y1,
 double x2, double y2)
{
  int triId = NTris;

  assert(NTris < MAX_TRIS);
  NTris++;

  Tri_coord0[triId][0] = x0;
  Tri_coord0[triId][1] = y0;

  Tri_coord1[triId][0] = x1;
  Tri_coord1[triId][1] = y1;

  Tri_coord2[triId][0] = x2;
  Tri_coord2[triId][1] = y2;

  return triId;
}

void WriteTriFile (const char *fileName)
{
  FILE *fp;
  int iTri;

  assert(0 < NTris);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  fprintf (fp, "%d\n", 
	   NTris);
  
  for (iTri = 0; iTri < NTris; iTri++) {
    fprintf 
      (fp, "%d   %22.15e %22.15e   %22.15e %22.15e   %22.15e %22.15e\n", 
       iTri,
       Tri_coord0[iTri][0], Tri_coord0[iTri][1],
       Tri_coord1[iTri][0], Tri_coord1[iTri][1],
       Tri_coord2[iTri][0], Tri_coord2[iTri][1]);
  }

  fclose (fp);
}



/************************
	 Command
*************************/

static int IsEndPoint
(double sx0, double sy0, 
 double sx1, double sy1,
 double x, double y)
{
  if (PointsAreSame (x, y,
		     sx0, sy0)) {
    return 1;
  }
  if (PointsAreSame (x, y,
		     sx1, sy1)) {
    return 1;
  }
  return 0;
}

static int CheckSeg 
(int segId,
 double sx0, double sy0, 
 double sx1, double sy1)
{
  int iSeg;

  if (Distance (sx0, sy0,
		sx1, sy1) < Eps) {
    return 0;
  }

  for (iSeg = 0; iSeg < NSegs; iSeg++) {
    if (iSeg != segId) {
      double x0 = Seg_coord0[iSeg][0];
      double y0 = Seg_coord0[iSeg][1]; 

      double x1 = Seg_coord1[iSeg][0];
      double y1 = Seg_coord1[iSeg][1]; 
      
      double intX, intY;
      int flag;

      flag = GetIntersectionBetweenSegs 
	(x0, y0, 
	 x1, y1,
	 sx0, sy0, 
	 sx1, sy1,
	 &intX, &intY);

      if (flag) {
	if (!IsEndPoint 
	    (x0, y0,
	     x1, y1,
	     intX, intY)) {
	  return 0;
	}
	if (!IsEndPoint 
	    (sx0, sy0,
	     sx1, sy1,
	     intX, intY)) {
	  return 0;
	}
      }
    }
  }

  return 1;
}

static void DoWithSeg (int segId)
{
  double sx0 = Seg_coord0[segId][0];
  double sy0 = Seg_coord0[segId][1]; 
  
  double sx1 = Seg_coord1[segId][0];
  double sy1 = Seg_coord1[segId][1]; 

  double a, b, c;
  double minDistFromSeg;
  int minSegId;
  int iSeg;

  GetLineEquation 
    (sx0, sy0,
     sx1, sy1,
     &a, &b, &c);

  minSegId = -1;
  minDistFromSeg = 0.0;
  for (iSeg = 0; iSeg < NSegs; iSeg++) {
    if (iSeg != segId) {
      double x = Seg_coord0[iSeg][0];
      double y = Seg_coord0[iSeg][1]; 
      
      double dist = a * x + b * y + c;
     
      if (dist < Eps) {
	continue;
      }

      if (CheckSeg (segId,
		    sx0, sy0, 
		    x, y)
	  && CheckSeg (segId,
		       sx1, sy1, 
		       x, y)) {
	if (minSegId == -1) {
	  minSegId = iSeg;
	  minDistFromSeg = dist;
	} else if (dist < minDistFromSeg) {
	  minSegId = iSeg;
	  minDistFromSeg = dist;
	}
      }
    }
  }
  assert(minSegId != -1);

  {
    double x = Seg_coord0[minSegId][0];
    double y = Seg_coord0[minSegId][1]; 

    int segId0 = FindSeg (sx0, sy0,
			  x, y);
    int segId1 = FindSeg (sx1, sy1,
			  x, y);
    
    if (segId0 == -1) {
      AddNewSeg (sx0, sy0,
		 x, y);
    } else {
      RemoveSeg (sx0, sy0,
		 x, y);
    }
    
    if (segId1 == -1) {
      AddNewSeg (x, y,
		 sx1, sy1);
    } else {
      RemoveSeg (x, y,
		 sx1, sy1);
    }
    
    RemoveSeg (sx0, sy0, 
	       sx1, sy1);
    
    AddNewTri (sx0, sy0,
	       sx1, sy1,
	       x, y);
    
  }
}

void PerformCommand (void)
{
  for (;;) {
    int segId = GetSeg ();

    if (segId == -1) break;

    DoWithSeg (segId);
  }
}



int main (int argc, char *argv[])
{
  if (argc != 3) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautoage_h_gm2d_tessellate  seg_file  \n"
       "        tri_file_OUT \n");
    exit (1);
  }

  ReadSegFile (argv[1]);

  PerformCommand ();

  WriteTriFile (argv[2]);

  return 0;
}
