/* advautoage_h_gm3d_polygon_view.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autogl.h>



/************************
	 Basic
*************************/

double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}

void GetNormalizedVector 
(double x, double y, double z,
 double *nx_OUT, double *ny_OUT, double *nz_OUT)
{
  double length = sqrt (x * x + y * y + z * z);

  *nx_OUT = x / length;
  *ny_OUT = y / length;
  *nz_OUT = z / length;
}

void GetVectorProduct 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  *x_OUT = y0 * z1 - z0 * y1;
  *y_OUT = z0 * x1 - x0 * z1;
  *z_OUT = x0 * y1 - y0 * x1;
}



/********************************
         Polygon Data
********************************/

/* Polygon */

#define MAX_POLYGONS 3000
#define MAX_POLYGON_LOOPS 100
#define MAX_POLYGON_VERTEXS 100

int NPolygons;
int Polygon_nLoops[MAX_POLYGONS];
int Polygon_nVertexs[MAX_POLYGONS][MAX_POLYGON_LOOPS];
double Polygon_vertexs[MAX_POLYGONS][MAX_POLYGON_LOOPS][MAX_POLYGON_VERTEXS][3];



void ReadPolygonFile (const char *fileName)
{
  FILE *fp;
  int iPolygon;
  int iLoop;
  int iVertex;

  assert(NPolygons == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NPolygons);
  assert(0 <= NPolygons);
  assert(NPolygons <= MAX_POLYGONS);
  
  for (iPolygon = 0; iPolygon < NPolygons; iPolygon++) {
    int polygonId;
    int nLoops;
    
    fscanf (fp, "%d %d", 
	    &polygonId,
	    &nLoops);
    assert(polygonId == iPolygon);
    assert(0 < nLoops);
    
    Polygon_nLoops[iPolygon] = nLoops;
    
    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs;

      fscanf (fp, "%d", 
	      &nVertexs);
      assert(2 < nVertexs);
      
      Polygon_nVertexs[iPolygon][iLoop] = nVertexs;

      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	double x, y, z;

	fscanf (fp, "%lf %lf %lf", 
		&x, &y, &z);
	
	Polygon_vertexs[iPolygon][iLoop][iVertex][0] = x;
	Polygon_vertexs[iPolygon][iLoop][iVertex][1] = y; 
	Polygon_vertexs[iPolygon][iLoop][iVertex][2] = z; 
      }
    }
  }

  fclose (fp);
}

void Polygon_getRange
(double *minRangeX_OUT, double *minRangeY_OUT, double *minRangeZ_OUT,
 double *maxRangeX_OUT, double *maxRangeY_OUT, double *maxRangeZ_OUT)
{
  int iPolygon;
  int iLoop;
  int iVertex;
  
  assert(0 < NPolygons);
  
  *minRangeX_OUT = *maxRangeX_OUT = Polygon_vertexs[0][0][0][0];
  *minRangeY_OUT = *maxRangeY_OUT = Polygon_vertexs[0][0][0][1];
  *minRangeZ_OUT = *maxRangeZ_OUT = Polygon_vertexs[0][0][0][2];
  
  for (iPolygon = 0; iPolygon < NPolygons; iPolygon++) {
    int nLoops = Polygon_nLoops[iPolygon];
    
    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs = Polygon_nVertexs[iPolygon][iLoop];
      
      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	double x = Polygon_vertexs[iPolygon][iLoop][iVertex][0];
	double y = Polygon_vertexs[iPolygon][iLoop][iVertex][1];
	double z = Polygon_vertexs[iPolygon][iLoop][iVertex][2];
	
	if (x < *minRangeX_OUT) *minRangeX_OUT = x;
	if (y < *minRangeY_OUT) *minRangeY_OUT = y;
	if (z < *minRangeZ_OUT) *minRangeZ_OUT = z;
	
	if (*maxRangeX_OUT < x) *maxRangeX_OUT = x;
	if (*maxRangeY_OUT < y) *maxRangeY_OUT = y;
	if (*maxRangeZ_OUT < z) *maxRangeZ_OUT = z;
      }
    }
  }
}

void Polygon_getNormal
(int polygonId,
 double *nx_OUT, double *ny_OUT, double *nz_OUT)
{
  const int outerLoopId = 0;
  int nVertexs;
  double x0, y0, z0;
  double x1, y1, z1;
  double x2, y2, z2;
  double nx, ny, nz;

  assert(0 <= polygonId);
  assert(polygonId < NPolygons);

  nVertexs = Polygon_nVertexs[polygonId][outerLoopId];

  x0 = Polygon_vertexs[polygonId][outerLoopId][0][0];
  y0 = Polygon_vertexs[polygonId][outerLoopId][0][1];
  z0 = Polygon_vertexs[polygonId][outerLoopId][0][2];

  x1 = Polygon_vertexs[polygonId][outerLoopId][1][0];
  y1 = Polygon_vertexs[polygonId][outerLoopId][1][1];
  z1 = Polygon_vertexs[polygonId][outerLoopId][1][2];

  x2 = Polygon_vertexs[polygonId][outerLoopId][nVertexs - 1][0];
  y2 = Polygon_vertexs[polygonId][outerLoopId][nVertexs - 1][1];
  z2 = Polygon_vertexs[polygonId][outerLoopId][nVertexs - 1][2];

  GetVectorProduct 
    (x1 - x0, y1- y0, z1 - z0,
     x2 - x0, y2- y0, z2 - z0,
     &nx, &ny, &nz);
  GetNormalizedVector
    (nx, ny, nz,
     &nx, &ny, &nz);

  *nx_OUT = nx;
  *ny_OUT = ny;
  *nz_OUT = nz;
}



/********************************
	 Visualization
********************************/

int Vertex = 1;
int Edge = 1;
int Normal = 1;
double Ratio = 0.9;
double ArrowScale = 1.0;

void SetView (void)
{
  double minRangeX, minRangeY, minRangeZ;
  double maxRangeX, maxRangeY, maxRangeZ;
  double centerX, centerY, centerZ;
  double size;
  
  Polygon_getRange 
    (&minRangeX, &minRangeY, &minRangeZ,
     &maxRangeX, &maxRangeY, &maxRangeZ);
  centerX = (minRangeX + maxRangeX) / 2;
  centerY = (minRangeY + maxRangeY) / 2;
  centerZ = (minRangeZ + maxRangeZ) / 2;
  size = Distance (minRangeX, minRangeY, minRangeZ,
		   maxRangeX, maxRangeY, maxRangeZ) / 2;

  AutoGL_SetViewSize (size);
  AutoGL_SetViewCenter (centerX, centerY, centerZ); 
}

void DrawPolygon (void) 
{
  int iPolygon;
  int iLoop;
  int iVertex;

  for (iPolygon = 0; iPolygon < NPolygons; iPolygon++) {
    int nLoops = Polygon_nLoops[iPolygon];

    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs = Polygon_nVertexs[iPolygon][iLoop];
      double centerX, centerY, centerZ;
      
      centerX = 0.0;
      centerY = 0.0;
      centerZ = 0.0;
      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	double x = Polygon_vertexs[iPolygon][iLoop][iVertex][0];
	double y = Polygon_vertexs[iPolygon][iLoop][iVertex][1];
	double z = Polygon_vertexs[iPolygon][iLoop][iVertex][2];
	
	centerX += x;
	centerY += y;
	centerZ += z;
      }
      centerX /= nVertexs;
      centerY /= nVertexs;
      centerZ /= nVertexs;

      if (Vertex) {
	AutoGL_SetColor (1, 0, 0);
	
	for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	  double x = Polygon_vertexs[iPolygon][iLoop][iVertex][0];
	  double y = Polygon_vertexs[iPolygon][iLoop][iVertex][1];
	  double z = Polygon_vertexs[iPolygon][iLoop][iVertex][2];

	  double vx = centerX + (x - centerX) * Ratio;
	  double vy = centerY + (y - centerY) * Ratio;
	  double vz = centerZ + (z - centerZ) * Ratio;

	  AutoGL_DrawCrossMark3D (vx, vy, vz,
				  3);
	}
      }

      if (Edge) {
	AutoGL_SetColor (0, 1, 0);
	
	for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	  int v0 = iVertex;
	  int v1 = (iVertex + 1) % nVertexs;

	  double x0 = Polygon_vertexs[iPolygon][iLoop][v0][0];
	  double y0 = Polygon_vertexs[iPolygon][iLoop][v0][1];
	  double z0 = Polygon_vertexs[iPolygon][iLoop][v0][2];

	  double x1 = Polygon_vertexs[iPolygon][iLoop][v1][0];
	  double y1 = Polygon_vertexs[iPolygon][iLoop][v1][1];
	  double z1 = Polygon_vertexs[iPolygon][iLoop][v1][2];

	  double vx0 = centerX + (x0 - centerX) * Ratio;
	  double vy0 = centerY + (y0 - centerY) * Ratio;
	  double vz0 = centerZ + (z0 - centerZ) * Ratio;
	  
	  double vx1 = centerX + (x1 - centerX) * Ratio;
	  double vy1 = centerY + (y1 - centerY) * Ratio;
	  double vz1 = centerZ + (z1 - centerZ) * Ratio;
	  
	  AutoGL_DrawLine (vx0, vy0, vz0,
			   vx1, vy1, vz1);
	}
      }
    }
  }
}

void DrawPolygonNormal (void) 
{
  int iPolygon;
  int iVertex;

  AutoGL_SetColor (1, 1, 1);

  for (iPolygon = 0; iPolygon < NPolygons; iPolygon++) {
    const int outerLoopId = 0;
    int nVertexs = Polygon_nVertexs[iPolygon][outerLoopId];
    double nx, ny, nz;
    double centerX, centerY, centerZ;
    double vx0, vy0, vz0;
    double vx, vy, vz;

    Polygon_getNormal (iPolygon,
		       &nx, &ny, &nz);

    centerX = 0.0;
    centerY = 0.0;
    centerZ = 0.0;
    for (iVertex = 0; iVertex < nVertexs; iVertex++) {
      double x = Polygon_vertexs[iPolygon][outerLoopId][iVertex][0];
      double y = Polygon_vertexs[iPolygon][outerLoopId][iVertex][1];
      double z = Polygon_vertexs[iPolygon][outerLoopId][iVertex][2];
      
      centerX += x;
      centerY += y;
      centerZ += z;
    }
    centerX /= nVertexs;
    centerY /= nVertexs;
    centerZ /= nVertexs;
    
    vx0 = Polygon_vertexs[iPolygon][outerLoopId][0][0];
    vy0 = Polygon_vertexs[iPolygon][outerLoopId][0][1];
    vz0 = Polygon_vertexs[iPolygon][outerLoopId][0][2];

    vx = centerX + (vx0 - centerX) * Ratio;
    vy = centerY + (vy0 - centerY) * Ratio;
    vz = centerZ + (vz0 - centerZ) * Ratio;

    AutoGL_DrawConeArrow3D 
      (vx, vy, vz,
       vx + nx * ArrowScale,
       vy + ny * ArrowScale,
       vz + nz * ArrowScale,
       0.3);
  }
}

void RedrawView (void) 
{
  AutoGL_DrawCoordinateSystem3D (20);

  DrawPolygon ();

  if (Normal) {
    DrawPolygonNormal ();
  }
}

void UpdateModel (void) 
{
  AutoGL_DrawView ();
}



void AutoGL_SetUp (int argc, char *argv[]) 
{
  if (argc != 2) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautoage_h_gm3d_polygon_view  polygon_file  \n");
    exit (1);
  }

  ReadPolygonFile (argv[1]);

  SetView ();
  
  AutoGL_SetViewRedrawCallback (RedrawView);

  AutoGL_SetDefaultCallbackInMode3D (NULL);
  AutoGL_EnableDragInMode3D ();



  AutoGL_AddGroup ("model");    
  AutoGL_SetPanelInMode3D ();

  AutoGL_AddComment ();
  AutoGL_SetLabel ("Model Parameters");
  
  AutoGL_AddBoolean (&Vertex, "Vertex");
  AutoGL_AddBoolean (&Edge, "Edge");
  AutoGL_AddReal (&Ratio, "Ratio");
  AutoGL_AddBoolean (&Normal, "Normal");
  AutoGL_AddReal (&ArrowScale, "ArrowScale");

  AutoGL_AddCallback (UpdateModel, "UpdateModel");      
}
