/* advautoage_h_gm3d_tri_inout3.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include <autores.h>



#if 1
#define PRINT_POINT_ID
#endif



/************************
	 Basic
*************************/

#define DEFAULT_EPS 0.00001
#define AREA_COORD_EPS 0.00001

double Eps = DEFAULT_EPS;



double Distance2D (double x0, double y0,
		   double x1, double y1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0));
}
void GetNormalizedVector 
(double x, double y, double z,
 double *nx_OUT, double *ny_OUT, double *nz_OUT)
{
  double length = sqrt (x * x + y * y + z * z);

  *nx_OUT = x / length;
  *ny_OUT = y / length;
  *nz_OUT = z / length;
}

void GetVectorProduct 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  *x_OUT = y0 * z1 - z0 * y1;
  *y_OUT = z0 * x1 - x0 * z1;
  *z_OUT = x0 * y1 - y0 * x1;
}

void GetPlaneEquation 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double x2, double y2, double z2,
 double *a_OUT, double *b_OUT, double *c_OUT, double *d_OUT)
{
  double nx, ny, nz;

  GetVectorProduct (x1 - x0, y1 - y0, z1 - z0,
		    x2 - x0, y2 - y0, z2 - z0,
		    &nx, &ny, &nz);
  GetNormalizedVector (nx, ny, nz,
		       &nx, &ny, &nz);
  *a_OUT = nx;
  *b_OUT = ny;
  *c_OUT = nz;
  *d_OUT = -(nx * x0 + ny * y0 + nz * z0);
}

double SignedTriArea2D
(double x0, double y0,
 double x1, double y1,
 double x2, double y2)
{
  return ((x1 - x0) * (y2 - y0) - (y1 - y0) * (x2 - x0)) / 2;
}

int TriContainsPoint2D 
(double x0, double y0,
 double x1, double y1,
 double x2, double y2,
 double x, double y)
{
  double area = SignedTriArea2D (x0, y0, 
				 x1, y1, 
				 x2, y2);
  double area0 = SignedTriArea2D (x1, y1, 
				  x2, y2, 
				  x, y);
  double area1 = SignedTriArea2D (x2, y2, 
				  x0, y0, 
				  x, y);
  double area2 = SignedTriArea2D (x0, y0, 
				  x1, y1, 
				  x, y);
  
  double L0 = area0 / area;
  double L1 = area1 / area;
  double L2 = area2 / area;

  if (0.0 - AREA_COORD_EPS < L0 
      && L0 < 1.0 + AREA_COORD_EPS
      && 0.0 - AREA_COORD_EPS < L1 
      && L1 < 1.0 + AREA_COORD_EPS
      && 0.0 - AREA_COORD_EPS < L2 
      && L2 < 1.0 + AREA_COORD_EPS) {
    return 1;
  }
  return 0;
}

double TriArea
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double x2, double y2, double z2)
{
  double vx, vy, vz;
  
  GetVectorProduct (x1 - x0, y1 - y0, z1 - z0,
		    x2 - x0, y2 - y0, z2 - z0,
		    &vx, &vy, &vz);
  return sqrt (vx * vx + vy * vy + vz * vz) / 2;
}

int TriContainsPoint 
(double tx0, double ty0, double tz0, 
 double tx1, double ty1, double tz1, 
 double tx2, double ty2, double tz2, 
 double x, double y, double z)
{
  double area = TriArea (tx0, ty0, tz0, 
			 tx1, ty1, tz1, 
			 tx2, ty2, tz2);
  double area0 = TriArea (tx1, ty1, tz1, 
			  tx2, ty2, tz2, 
			  x, y, z);
  double area1 = TriArea (tx2, ty2, tz2, 
			  tx0, ty0, tz0, 
			  x, y, z);
  double area2 = TriArea (tx0, ty0, tz0, 
			  tx1, ty1, tz1, 
			  x, y, z);
  
  if (fabs ((area0 + area1 + area2) / area - 1.0) < AREA_COORD_EPS) {
    return 1;
  }
  return 0;
}



/************************
        Triangles
*************************/

/* Triangle */

#define MAX_TRIS 100000

int NTris;
double Tri_coord0[MAX_TRIS][3];
double Tri_coord1[MAX_TRIS][3];
double Tri_coord2[MAX_TRIS][3];

double Tri_planeA[MAX_TRIS];
double Tri_planeB[MAX_TRIS];
double Tri_planeC[MAX_TRIS];
double Tri_planeD[MAX_TRIS];

double Tri_min[MAX_TRIS][3];
double Tri_max[MAX_TRIS][3];

double Tri_minRange[3];
double Tri_maxRange[3];

AutoRes_Bucket2D TriBucket;



void ReadTriFile (const char *fileName)
{
  FILE *fp;
  int iTri;

  assert(NTris == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NTris);
  assert(0 <= NTris);
  assert(NTris <= MAX_TRIS);
  
  for (iTri = 0; iTri < NTris; iTri++) {
    int triId;
    double x0, y0, z0, x1, y1, z1, x2, y2, z2;

    fscanf (fp, "%d %lf %lf %lf %lf %lf %lf %lf %lf %lf", 
	    &triId,
	    &x0, &y0, &z0, 
	    &x1, &y1, &z1, 
	    &x2, &y2, &z2);
    assert(triId == iTri);

    Tri_coord0[iTri][0] = x0;
    Tri_coord0[iTri][1] = y0; 
    Tri_coord0[iTri][2] = z0; 

    Tri_coord1[iTri][0] = x1;
    Tri_coord1[iTri][1] = y1; 
    Tri_coord1[iTri][2] = z1; 

    Tri_coord2[iTri][0] = x2;
    Tri_coord2[iTri][1] = y2; 
    Tri_coord2[iTri][2] = z2; 
  }

  fclose (fp);
}

void PrepareTriData (void)
{
  int iTri;

  for (iTri = 0; iTri < NTris; iTri++) {
    double tx0 = Tri_coord0[iTri][0];
    double ty0 = Tri_coord0[iTri][1];
    double tz0 = Tri_coord0[iTri][2];
    
    double tx1 = Tri_coord1[iTri][0];
    double ty1 = Tri_coord1[iTri][1];
    double tz1 = Tri_coord1[iTri][2];
    
    double tx2 = Tri_coord2[iTri][0];
    double ty2 = Tri_coord2[iTri][1];
    double tz2 = Tri_coord2[iTri][2];

    {
      double a, b, c, d;
      
      GetPlaneEquation 
	(tx0, ty0, tz0,
	 tx1, ty1, tz1,
	 tx2, ty2, tz2,
	 &a, &b, &c, &d);

      Tri_planeA[iTri] = a;
      Tri_planeB[iTri] = b;
      Tri_planeC[iTri] = c;
      Tri_planeD[iTri] = d;
    }

    {
      double minX = tx0;
      double minY = ty0;
      double minZ = tz0;
      
      double maxX = tx0;
      double maxY = ty0;
      double maxZ = tz0;
      
      if (tx1 < minX) {
	minX = tx1;
      }
      if (ty1 < minY) {
	minY = ty1;
      }
      if (tz1 < minZ) {
	minZ = tz1;
      }

      if (maxX < tx1) {
	maxX = tx1;
      }
      if (maxY < ty1) {
	maxY = ty1;
      }
      if (maxZ < tz1) {
	maxZ = tz1;
      }

      if (tx2 < minX) {
	minX = tx2;
      }
      if (ty2 < minY) {
	minY = ty2;
      }
      if (tz2 < minZ) {
	minZ = tz2;
      }

      if (maxX < tx2) {
	maxX = tx2;
      }
      if (maxY < ty2) {
	maxY = ty2;
      }
      if (maxZ < tz2) {
	maxZ = tz2;
      }

      Tri_min[iTri][0] = minX;
      Tri_min[iTri][1] = minY;
      Tri_min[iTri][2] = minZ;

      Tri_max[iTri][0] = maxX;
      Tri_max[iTri][1] = maxY;
      Tri_max[iTri][2] = maxZ;
    }
  }

  Tri_minRange[0] = Tri_min[0][0];
  Tri_minRange[1] = Tri_min[0][1];
  Tri_minRange[2] = Tri_min[0][2];

  Tri_maxRange[0] = Tri_max[0][0];
  Tri_maxRange[1] = Tri_max[0][1];
  Tri_maxRange[2] = Tri_max[0][2];

  for (iTri = 0; iTri < NTris; iTri++) {
    double minX = Tri_min[iTri][0];
    double minY = Tri_min[iTri][1];
    double minZ = Tri_min[iTri][2];

    double maxX = Tri_max[iTri][0];
    double maxY = Tri_max[iTri][1];
    double maxZ = Tri_max[iTri][2];

    if (minX < Tri_minRange[0]) {
      Tri_minRange[0] = minX;
    }
    if (minY < Tri_minRange[1]) {
      Tri_minRange[1] = minY;
    }
    if (minZ < Tri_minRange[2]) {
      Tri_minRange[2] = minZ;
    }

    if (Tri_maxRange[0] < maxX) {
      Tri_maxRange[0] = maxX;
    }
    if (Tri_maxRange[1] < maxY) {
      Tri_maxRange[1] = maxY;
    }
    if (Tri_maxRange[2] < maxZ) {
      Tri_maxRange[2] = maxZ;
    }
  }
}

double Tri_containsPointOnXYPlane 
(int triId, 
 double px, double py, double pz)
{
  double tx0 = Tri_coord0[triId][0];
  double ty0 = Tri_coord0[triId][1];
  
  double tx1 = Tri_coord1[triId][0];
  double ty1 = Tri_coord1[triId][1];
  
  double tx2 = Tri_coord2[triId][0];
  double ty2 = Tri_coord2[triId][1];
  
  return TriContainsPoint2D
    (tx0, ty0,
     tx1, ty1,
     tx2, ty2,
     px, py);
}

void Tri_createBucket (void)
{
  double minRangeX, minRangeY;
  double maxRangeX, maxRangeY;
  double bucketSize;
  int iTri;
  int xDivisions, yDivisions;
  int nOffsetDivisions = 23;
  double offsetX, offsetY;
  int nAverageLinksPerCell = 1000;

  assert(0 < NTris);

  minRangeX = Tri_minRange[0];
  minRangeY = Tri_minRange[1];

  maxRangeX = Tri_maxRange[0];
  maxRangeY = Tri_maxRange[1];

  bucketSize = 0.0;
  for (iTri = 0; iTri < NTris; iTri++) {
    double tx0 = Tri_coord0[iTri][0];
    double ty0 = Tri_coord0[iTri][1];
    
    double tx1 = Tri_coord1[iTri][0];
    double ty1 = Tri_coord1[iTri][1];
    
    double tx2 = Tri_coord2[iTri][0];
    double ty2 = Tri_coord2[iTri][1];
    
    double dist01 = Distance2D (tx0, ty0, 
				tx1, ty1);
    double dist12 = Distance2D (tx1, ty1, 
				tx2, ty2);
    double dist20 = Distance2D (tx2, ty2, 
				tx0, ty0);
    bucketSize += (dist01 + dist12 + dist20) / 3;
  }
  bucketSize /= NTris;

  /* NEED TO REDEFINE */
  bucketSize *= 1.0;

  xDivisions = (maxRangeX - minRangeX) / bucketSize;
  if (xDivisions < 1) xDivisions = 1;
  yDivisions = (maxRangeY - minRangeY) / bucketSize;
  if (yDivisions < 1) yDivisions = 1;

  if (xDivisions < 5) xDivisions = 5;
  if (yDivisions < 5) yDivisions = 5;

  offsetX = (maxRangeX - minRangeX) / nOffsetDivisions;
  offsetY = (maxRangeY - minRangeY) / nOffsetDivisions;

  AutoRes_Bucket2D_initialize 
    (&TriBucket,
     minRangeX - offsetX, minRangeY - offsetY,
     maxRangeX + offsetX, maxRangeY + offsetY,
     xDivisions, yDivisions, 
     nAverageLinksPerCell);

  for (iTri = 0; iTri < NTris; iTri++) {
    double minRangeX = Tri_min[iTri][0];
    double minRangeY = Tri_min[iTri][1];
    
    double maxRangeX = Tri_max[iTri][0];
    double maxRangeY = Tri_max[iTri][1];
    
    AutoRes_Bucket2D_addItemWithin 
      (&TriBucket,
       minRangeX, minRangeY,
       maxRangeX, maxRangeY,
       iTri);
  }



  fprintf (stderr, " xDivisions %d yDivisions %d \n",
	   xDivisions, yDivisions);
}

void Tri_deleteBucket (void)
{
  AutoRes_Bucket2D_terminate 
    (&TriBucket);
}



#define RAY_UNKNOWN      0
#define RAY_INTERSECT    1
#define RAY_ON_TRIANGLE  2

static int CheckTriAndPointAlongZRay 
(int triId,
 double px, double py, double pz,
 double *intersectionZ_OUT)
{
  double tx0 = Tri_coord0[triId][0];
  double ty0 = Tri_coord0[triId][1];
  double tz0 = Tri_coord0[triId][2];
  
  double tx1 = Tri_coord1[triId][0];
  double ty1 = Tri_coord1[triId][1];
  double tz1 = Tri_coord1[triId][2];
  
  double tx2 = Tri_coord2[triId][0];
  double ty2 = Tri_coord2[triId][1];
  double tz2 = Tri_coord2[triId][2];

  double minX = Tri_min[triId][0];
  double minY = Tri_min[triId][1];
  double minZ = Tri_min[triId][2];

  double maxX = Tri_max[triId][0];
  double maxY = Tri_max[triId][1];
  double maxZ = Tri_max[triId][2];

  double a = Tri_planeA[triId];
  double b = Tri_planeB[triId];
  double c = Tri_planeC[triId];
  double d = Tri_planeD[triId];
  double signedDist = a * px + b * py + c * pz + d;

  int inside3D;
  
  if (fabs (signedDist) < Eps) {
    if (minX - Eps < px && px < maxX + Eps
	&& minY - Eps < py && py < maxY + Eps
	&& minZ - Eps < pz && pz < maxZ + Eps) {
      inside3D = TriContainsPoint
	(tx0, ty0, tz0,
	 tx1, ty1, tz1,
	 tx2, ty2, tz2,
	 px, py, pz);
    } else {
      inside3D = 0;
    }
  } else {
    inside3D = 0;
  }
  
  if (inside3D) {
    return RAY_ON_TRIANGLE;

  } else {
    if (fabs (c) < Eps) {
      /* parallel to Z axis */
      
      return RAY_UNKNOWN;
    } else {
      int inside2D;

      if (minX - Eps < px && px < maxX + Eps
	  && minY - Eps < py && py < maxY + Eps) {
	inside2D = TriContainsPoint2D
	  (tx0, ty0,
	   tx1, ty1,
	   tx2, ty2,
	   px, py);
      } else {
	inside2D = 0;
      }
      
      if (fabs (signedDist) < Eps) {
	/* point is on plane of triangle */
	/* on boundary, or unknown */
	
	if (inside2D) {
	  return RAY_ON_TRIANGLE;
	} else {
	  return RAY_UNKNOWN;
	}
	
      } else {
	/* point is out of plane of triangle */
	/* intersect or unknown */
	
	if (inside2D) {
	  *intersectionZ_OUT = - (d + a * px + b * py) / c;

	  if (pz < *intersectionZ_OUT) {
	    return RAY_INTERSECT;
	  } else {
	    assert(*intersectionZ_OUT < pz);

	    return RAY_UNKNOWN;
	  }
	} else {
	  return RAY_UNKNOWN;
	}
	
      }
    }
  }

  assert(0);

  return RAY_UNKNOWN;
}

#define MAX_INTERSECTIONS 1000
int NIntersections;
double IntersectionZs[MAX_INTERSECTIONS];

static void ClearIntersection (void)
{
  NIntersections = 0;
}

static void AddIntersection (double z)
{
  int iIntersection;
  int found;
  
  found = 0;
  for (iIntersection = 0; 
       iIntersection < NIntersections; 
       iIntersection++) {
    if (fabs (IntersectionZs[iIntersection] - z) < Eps) {
      found = 1;
      break;
    }
  }
  
  if (!found) {
    assert(NIntersections < MAX_INTERSECTIONS);
    
    IntersectionZs[NIntersections] = z;
    NIntersections++;
  }
}

#define POINT_INOUT_OUTSIDE      0
#define POINT_INOUT_INSIDE       1
#define POINT_INOUT_ON_BOUNDARY  2

int Tri_checkInout (double px, double py, double pz)
{  
  int cx, cy;
  int linkIndex;

  ClearIntersection ();

  AutoRes_Bucket2D_getIndexAt 
    (&TriBucket, 
     px, py, 
     &cx, &cy);

  if (AutoRes_Bucket2D_isOutOfRange 
      (&TriBucket, 
       cx, cy)) {
    linkIndex = AutoRes_Bucket2D_startLinkIndexOutOfRange 
      (&TriBucket);
  } else {
    linkIndex = AutoRes_Bucket2D_startLinkIndexIn 
      (&TriBucket, 
       cx, cy);
  }

  for (/* nothing */;
		    linkIndex != -1;
		    linkIndex = AutoRes_Bucket2D_nextLinkIndex 
		      (&TriBucket, 
		       linkIndex)) {
    int triId = AutoRes_Bucket2D_itemId 
      (&TriBucket, 
       linkIndex);
    
    double intersectionZ;
    int checkResult = CheckTriAndPointAlongZRay 
      (triId, 
       px, py, pz,
       &intersectionZ);
    
    switch (checkResult) {
    case RAY_UNKNOWN:
      break;
    case RAY_INTERSECT:
      AddIntersection (intersectionZ);
      break;
    case RAY_ON_TRIANGLE:
      return POINT_INOUT_ON_BOUNDARY;
    default:
      assert(0);
      break;
    }
  }

  if (NIntersections % 2 == 1) {
    return POINT_INOUT_INSIDE;
  } else {
    return POINT_INOUT_OUTSIDE;
  }
}



/************************
        Points
*************************/

/* Point */

#define MAX_POINTS 5000000


int NPoints;
double Point_coord[MAX_POINTS][3];
int Point_inoutType[MAX_POINTS];



void ReadPointFile (const char *fileName)
{
  FILE *fp;
  int iPoint;
  
  assert(NPoints == 0);
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NPoints);
  assert(0 <= NPoints);
  assert(NPoints <= MAX_POINTS);
  
  for (iPoint = 0; iPoint < NPoints; iPoint++) {
    int pointId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &pointId,
	    &x, &y, &z);
    assert(pointId == iPoint);
    Point_coord[iPoint][0] = x;
    Point_coord[iPoint][1] = y; 
    Point_coord[iPoint][2] = z; 
  }

  fclose (fp);
}

void WritePointInoutFile (const char *fileName)
{
  FILE *fp;
  int iPoint;
  
  assert(0 < NPoints);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n", 
	   NPoints);
  
  for (iPoint = 0; iPoint < NPoints; iPoint++) {
    fprintf (fp, "%d   %d\n", 
	     iPoint, Point_inoutType[iPoint]);
  }

  fclose (fp);
}



/************************
	 Command
*************************/

void PerformCommand (void)
{
  int iPoint;
  
  PrepareTriData ();
  Tri_createBucket ();

  for (iPoint = 0; iPoint < NPoints; iPoint++) {
    double px = Point_coord[iPoint][0];
    double py = Point_coord[iPoint][1];
    double pz = Point_coord[iPoint][2];
    
    int inoutType0 = Tri_checkInout (px, py, pz);

    if (inoutType0 == POINT_INOUT_ON_BOUNDARY) {
      Point_inoutType[iPoint] = POINT_INOUT_ON_BOUNDARY;

    } else {
      double delta = Eps * 10;
      int inoutType1 = Tri_checkInout (px - delta, py, pz);
      int inoutType2 = Tri_checkInout (px, py - delta, pz);

      if (inoutType1 == inoutType0
	  && inoutType2 == inoutType0) {
	Point_inoutType[iPoint] = inoutType0;

      } else {
	int inoutType3 = Tri_checkInout (px + delta, py + delta, pz);
	int insideCount;
	
	insideCount = 0;
	if (inoutType1 == POINT_INOUT_INSIDE) {
	  insideCount++;
	}
	if (inoutType2 == POINT_INOUT_INSIDE) {
	  insideCount++;
	}
	if (inoutType3 == POINT_INOUT_INSIDE) {
	  insideCount++;
	}
	
	if (2 <= insideCount) {
	  Point_inoutType[iPoint] = POINT_INOUT_INSIDE;
	} else {
	  Point_inoutType[iPoint] = POINT_INOUT_OUTSIDE;
	}
      }
    }

#ifdef PRINT_POINT_ID
    if (iPoint % 10000 == 0) {
      printf (" iPoint %d \n",
	      iPoint);
    }
#endif
  }

  Tri_deleteBucket ();
}



int main (int argc, char *argv[])
{
  if (argc != 4) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautoage_h_gm3d_tri_inout3  tri_file  \n"
       "        point_file \n"
       "        point_inout_file_OUT \n");
    exit (1);
  }

  ReadTriFile (argv[1]);
  ReadPointFile (argv[2]);

  PerformCommand ();

  WritePointInoutFile (argv[3]);

  return 0;
}
