/* advautoage_h_gm3d_attach_polygon.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

#define DEFAULT_EPS 0.00001

double Eps = DEFAULT_EPS;



int Equal (double value0, double value1) 
{
  return fabs (value1 - value0) < Eps;
}

int Between (double value, double range0, double range1) 
{
  return (range0 - Eps < value && value < range1 + Eps)
    || (range1 - Eps < value && value < range0 + Eps);
}

double Distance (double x0, double y0, double z0,
		 double x1, double y1, double z1) 
{
  return sqrt ((x1 - x0) * (x1 - x0) 
	       + (y1 - y0) * (y1 - y0)
	       + (z1 - z0) * (z1 - z0));
}

int PointsAreSame (double x0, double y0, double z0, 
		   double x1, double y1, double z1) 
{
  return Distance (x0, y0, z0,
		   x1, y1, z1) < Eps;
}

void GetNormalizedVector 
(double x, double y, double z,
 double *nx_OUT, double *ny_OUT, double *nz_OUT)
{
  double length = sqrt (x * x + y * y + z * z);

  *nx_OUT = x / length;
  *ny_OUT = y / length;
  *nz_OUT = z / length;
}

void GetVectorProduct 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  *x_OUT = y0 * z1 - z0 * y1;
  *y_OUT = z0 * x1 - x0 * z1;
  *z_OUT = x0 * y1 - y0 * x1;
}

void GetPlaneEquation 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double x2, double y2, double z2,
 double *a_OUT, double *b_OUT, double *c_OUT, double *d_OUT)
{
  double nx, ny, nz;

  GetVectorProduct (x1 - x0, y1 - y0, z1 - z0,
		    x2 - x0, y2 - y0, z2 - z0,
		    &nx, &ny, &nz);
  GetNormalizedVector (nx, ny, nz,
		       &nx, &ny, &nz);
  *a_OUT = nx;
  *b_OUT = ny;
  *c_OUT = nz;
  *d_OUT = -(nx * x0 + ny * y0 + nz * z0);
}

void GetAnyPerpendicularDirection
(double x, double y, double z,
 double *nx0_OUT, double *ny0_OUT, double *nz0_OUT,
 double *nx1_OUT, double *ny1_OUT, double *nz1_OUT)
{
  double length;

  GetVectorProduct (x, y, z,
		    1, 0, 0,
		    nx0_OUT, ny0_OUT, nz0_OUT);
  length = sqrt (*nx0_OUT * *nx0_OUT
		 + *ny0_OUT * *ny0_OUT
		 + *nz0_OUT * *nz0_OUT);
  if (length == 0.0) {
    GetVectorProduct (x, y, z,
		      0, 1, 0,
		      nx0_OUT, ny0_OUT, nz0_OUT);
  }
  GetNormalizedVector (*nx0_OUT, *ny0_OUT, *nz0_OUT,
		       nx0_OUT, ny0_OUT, nz0_OUT);
  GetVectorProduct (x, y, z,
		    *nx0_OUT, *ny0_OUT, *nz0_OUT,
		    nx1_OUT, ny1_OUT, nz1_OUT);
  GetNormalizedVector (*nx1_OUT, *ny1_OUT, *nz1_OUT,
		       nx1_OUT, ny1_OUT, nz1_OUT);
}

void GetProjectedPointOnPlane
(double a, double b, double c, double d, 
 double x, double y, double z,
 double *px_OUT, double *py_OUT, double *pz_OUT)
{
  double value = a * x + b * y + c * z + d;
  
  *px_OUT = x - a * value;
  *py_OUT = y - b * value;
  *pz_OUT = z - c * value;
}

int PlanesAreSameSide
(double p0_a, double p0_b, double p0_c, double p0_d,
 double p1_a, double p1_b, double p1_c, double p1_d)
{
  if (Equal (p0_a, p1_a)
      && Equal (p0_b, p1_b)
      && Equal (p0_c, p1_c)
      && Equal (p0_d, p1_d)) {
    return 1;
  }
  return 0;
}

int PlanesAreOppositeSide
(double p0_a, double p0_b, double p0_c, double p0_d,
 double p1_a, double p1_b, double p1_c, double p1_d)
{
  if (Equal (p0_a, -p1_a)
      && Equal (p0_b, -p1_b)
      && Equal (p0_c, -p1_c)
      && Equal (p0_d, -p1_d)) {
    return 1;
  }
  return 0;
}

int GetIntersectionBetweenPlaneAndSeg 
(double a, double b, double c, double d, 
 double sx0, double sy0, double sz0, 
 double sx1, double sy1, double sz1, 
 double *intX_OUT, double *intY_OUT, double *intZ_OUT) 
{
  double value0 = a * sx0 + b * sy0 + c * sz0 + d;
  double value1 = a * sx1 + b * sy1 + c * sz1 + d;

  if (fabs (value0 - value1) < Eps) return 0;

  if (fabs (value0) < Eps) {
    *intX_OUT = sx0;
    *intY_OUT = sy0;
    *intZ_OUT = sz0;
    return 1;
  } else if (fabs (value1) < Eps) {
    *intX_OUT = sx1;
    *intY_OUT = sy1;
    *intZ_OUT = sz1;
    return 1;
  } else if (value0 * value1 < Eps * Eps) {
    *intX_OUT = (sx1 * value0 - sx0 * value1) / (value0 - value1);
    *intY_OUT = (sy1 * value0 - sy0 * value1) / (value0 - value1);
    *intZ_OUT = (sz1 * value0 - sz0 * value1) / (value0 - value1);
    return 1;
  }
  return 0;
}

double LineParameterAt 
(double originX, double originY, double originZ,
 double dirX, double dirY, double dirZ,
 double x, double y, double z)
{
  return (x - originX) * dirX
    + (y - originY) * dirY
    + (z - originZ) * dirZ;
}

double DistanceFromLine 
(double originX, double originY, double originZ,
 double dirX, double dirY, double dirZ,
 double x, double y, double z)
{
  double t = LineParameterAt (originX, originY, originZ,
			      dirX, dirY, dirZ,
			      x, y, z);
  
  return Distance (x, y, z,
		   originX + dirX * t, 
		   originY + dirY * t, 
		   originZ + dirZ * t);
}

int SegContainsPoint
(double sx0, double sy0, double sz0,
 double sx1, double sy1, double sz1,
 double px, double py, double pz)
{
  double originX, originY, originZ;
  double dirX, dirY, dirZ;
  double dist;

  originX = sx0;
  originY = sy0;
  originZ = sz0;
  GetNormalizedVector (sx1 - sx0, sy1 - sy0, sz1 - sz0,
		       &dirX, &dirY, &dirZ);
  
  dist = DistanceFromLine (originX, originY, originZ,
			   dirX, dirY, dirZ,
			   px, py, pz);
  if (dist < Eps) {
    double t0 = LineParameterAt (originX, originY, originZ,
				 dirX, dirY, dirZ,
				 sx0, sy0, sz0);
    double t1 = LineParameterAt (originX, originY, originZ,
				 dirX, dirY, dirZ,
				 sx1, sy1, sz1);
    double t = LineParameterAt (originX, originY, originZ,
				dirX, dirY, dirZ,
				px, py, pz);

    if (Between (t, t0, t1)) {
      return 1;
    }
  }
  return 0;
}



/***********************************
       2-D / 3-D Conversion
************************************/

static double Plane_a, Plane_b, Plane_c, Plane_d;
static double Plane_ox, Plane_oy, Plane_oz;
static double Plane_ux, Plane_uy, Plane_uz;
static double Plane_vx, Plane_vy, Plane_vz;

void SetUVCoordinate 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double x2, double y2, double z2)
{
  GetPlaneEquation (x0, y0, z0,
		    x1, y1, z1,
		    x2, y2, z2,
		    &Plane_a, &Plane_b, &Plane_c, &Plane_d);
  GetProjectedPointOnPlane
    (Plane_a, Plane_b, Plane_c, Plane_d,
     x0, y0, z0, 
     &Plane_ox, &Plane_oy, &Plane_oz);
  GetAnyPerpendicularDirection
    (Plane_a, Plane_b, Plane_c,
     &Plane_ux, &Plane_uy, &Plane_uz,
     &Plane_vx, &Plane_vy, &Plane_vz);
}

void GetUVOfCoordinate 
(double x, double y, double z,
 double *u_OUT, double *v_OUT)
{
  double u 
    = (x - Plane_ox) * Plane_ux
    + (y - Plane_oy) * Plane_uy
    + (z - Plane_oz) * Plane_uz;
  double v 
    = (x - Plane_ox) * Plane_vx
    + (y - Plane_oy) * Plane_vy
    + (z - Plane_oz) * Plane_vz;

  *u_OUT = u;
  *v_OUT = v;
}

void GetCoordinateOfUV 
(double u, double v,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  double x = Plane_ox + Plane_ux * u + Plane_vx * v;
  double y = Plane_oy + Plane_uy * u + Plane_vy * v;
  double z = Plane_oz + Plane_uz * u + Plane_vz * v;

  *x_OUT = x;
  *y_OUT = y;
  *z_OUT = z;
}



/********************************
         Polygon Data
********************************/

/* Polygon */

#define MAX_POLYGONS 1000
#define MAX_POLYGON_LOOPS 100
#define MAX_POLYGON_VERTEXS 100

int NPolygons;
int Polygon_nLoops[MAX_POLYGONS];
int Polygon_nVertexs[MAX_POLYGONS][MAX_POLYGON_LOOPS];
double Polygon_vertexs[MAX_POLYGONS][MAX_POLYGON_LOOPS][MAX_POLYGON_VERTEXS][3];

int Polygon_isA[MAX_POLYGONS];
int Polygon_planeId[MAX_POLYGONS];
int Polygon_isFrontSide[MAX_POLYGONS];

#define MAX_PLANES 1000
#define MAX_PLANE_POLYGONS 1000

int NPlanes;
int Plane_nPolygons[MAX_PLANES];
int Plane_polygonIds[MAX_PLANES][MAX_PLANE_POLYGONS];



static void ReadNewPolygon 
(FILE *fp, 
 int filePolygonId, 
 int polygonIsA)
{
  int polygonId = NPolygons;
  int nLoops;
  int iLoop;
  int iVertex;
    
  assert(NPolygons < MAX_POLYGONS);
  NPolygons++;
  
  Polygon_isA[polygonId] = polygonIsA;

  {
    int polygonIdFromFile;
    
    fscanf (fp, "%d %d", 
	    &polygonIdFromFile,
	    &nLoops);
    assert(polygonIdFromFile == filePolygonId);
    assert(0 < nLoops);
  
    Polygon_nLoops[polygonId] = nLoops;
  }
  
  for (iLoop = 0; iLoop < nLoops; iLoop++) {
    int nVertexs;
    
    fscanf (fp, "%d", 
	    &nVertexs);
    assert(2 < nVertexs);
    
    Polygon_nVertexs[polygonId][iLoop] = nVertexs;
    
    for (iVertex = 0; iVertex < nVertexs; iVertex++) {
      double x, y, z;
      
      fscanf (fp, "%lf %lf %lf", 
	      &x, &y, &z);
      
      Polygon_vertexs[polygonId][iLoop][iVertex][0] = x;
      Polygon_vertexs[polygonId][iLoop][iVertex][1] = y; 
      Polygon_vertexs[polygonId][iLoop][iVertex][2] = z;
    }
  }
}

void ReadPolygonFile 
(const char *fileNameA,
 const char *fileNameB)
{
  FILE *fp;
  int nPolygons;
  int iPolygon;
  int polygonIsA;

  assert(NPolygons == 0);



  fp = fopen (fileNameA, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &nPolygons);
  assert(0 <= nPolygons);
  
  polygonIsA = 1;
  for (iPolygon = 0; iPolygon < nPolygons; iPolygon++) {
    ReadNewPolygon (fp, iPolygon, polygonIsA);
  }

  fclose (fp);



  fp = fopen (fileNameB, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &nPolygons);
  assert(0 <= nPolygons);
  
  polygonIsA = 0;
  for (iPolygon = 0; iPolygon < nPolygons; iPolygon++) {
    ReadNewPolygon (fp, iPolygon, polygonIsA);
  }

  fclose (fp);
}

void Polygon_getPlaneEquation 
(int polygonId,
 double *a_OUT, double *b_OUT, double *c_OUT, double *d_OUT)
{
  int v0 = 0;
  int v1 = 1;
  int v2 = Polygon_nVertexs[polygonId][0] - 1;
  
  double x0 = Polygon_vertexs[polygonId][0][v0][0];
  double y0 = Polygon_vertexs[polygonId][0][v0][1];
  double z0 = Polygon_vertexs[polygonId][0][v0][2];
  
  double x1 = Polygon_vertexs[polygonId][0][v1][0];
  double y1 = Polygon_vertexs[polygonId][0][v1][1];
  double z1 = Polygon_vertexs[polygonId][0][v1][2];
  
  double x2 = Polygon_vertexs[polygonId][0][v2][0];
  double y2 = Polygon_vertexs[polygonId][0][v2][1];
  double z2 = Polygon_vertexs[polygonId][0][v2][2];
  
  GetPlaneEquation (x0, y0, z0,
		    x1, y1, z1,
		    x2, y2, z2,
		    a_OUT, b_OUT, c_OUT, d_OUT);
}



/************************
	 Command
*************************/

static void IdentifyPlane (void)
{
  int iPolygon;
  int iPlane;
  int count;

  assert(NPlanes == 0);

  for (iPolygon = 0; iPolygon < NPolygons; iPolygon++) {
    Polygon_planeId[iPolygon] = -1;
    Polygon_isFrontSide[iPolygon] = 0;
  }

  count = 0;
  for (;;) {
    int found;
    double a, b, c, d;

    found = 0;
    for (iPolygon = 0; iPolygon < NPolygons; iPolygon++) {
      if (Polygon_planeId[iPolygon] == -1) {
	found = 1;
	Polygon_getPlaneEquation (iPolygon,
				  &a, &b, &c, &d);
	
	Polygon_planeId[iPolygon] = count;
	Polygon_isFrontSide[iPolygon] = 1;
	break;
      }
    }

    if (!found) {
      break;
    }

    for (iPolygon = 0; iPolygon < NPolygons; iPolygon++) {
      if (Polygon_planeId[iPolygon] == -1) {
	double pa, pb, pc, pd;
	
	Polygon_getPlaneEquation (iPolygon,
				  &pa, &pb, &pc, &pd);
	
	if (PlanesAreSameSide (pa, pb, pc, pd,
			       a, b, c, d)) {
	  Polygon_planeId[iPolygon] = count;
	  Polygon_isFrontSide[iPolygon] = 1;
	} else if (PlanesAreOppositeSide (pa, pb, pc, pd,
					  a, b, c, d)) {
	  Polygon_planeId[iPolygon] = count;
	  Polygon_isFrontSide[iPolygon] = 0;
	}
      }
    }
    
    count++;
  }

  NPlanes = count;
  assert(0 < NPlanes);
  assert(NPlanes <= MAX_PLANES);

  for (iPlane = 0; iPlane < NPlanes; iPlane++) {
    Plane_nPolygons[iPlane] = 0;
  }

  for (iPolygon = 0; iPolygon < NPolygons; iPolygon++) {
    int planeId = Polygon_planeId[iPolygon];
    int planePolygonId;

    assert(0 <= planeId);

    planePolygonId = Plane_nPolygons[planeId];

    assert(Plane_nPolygons[planeId] < MAX_PLANE_POLYGONS);
    Plane_nPolygons[planeId]++;

    Plane_polygonIds[planeId][planePolygonId] = iPolygon;
  }
}



#define MAX_REGISTERS 10
#define MAX_REGISTER_POLYGONS 100

#define R0 0
#define R1 1
#define R2 2
#define R3 3
#define R4 4
#define R5 5
#define R6 6
#define R7 7
#define R8 8
#define R9 9

static int NRegisterPolygons[MAX_REGISTERS];
static int RegisterPolygon_nLoops[MAX_REGISTERS][MAX_REGISTER_POLYGONS];
static int RegisterPolygon_nVertexs
[MAX_REGISTERS][MAX_REGISTER_POLYGONS][MAX_POLYGON_LOOPS];
static double RegisterPolygon_vertexs
[MAX_REGISTERS][MAX_REGISTER_POLYGONS][MAX_POLYGON_LOOPS][MAX_POLYGON_VERTEXS][3];

#define SET_POLYGON_FILE_NAME_A "___attach_polygon__set2d_a.gm_polygon"
#define SET_POLYGON_FILE_NAME_B "___attach_polygon__set2d_b.gm_polygon"
#define SET_POLYGON_FILE_NAME_AB "___attach_polygon__set2d_ab.gm_polygon"

#define SET2D_COMMAND "advautoage_h_gm2d_set_polygon"

static void ClearRegisterPolygon
(int registerId)
{
  assert(0 <= registerId);
  assert(registerId < MAX_REGISTERS);
  
  NRegisterPolygons[registerId] = 0;
}

static void AddPolygonIntoRegisterPolygon
(int registerId,
 int polygonId)
{
  int registerPolygonId;
  int nLoops;
  int iLoop;
  int iVertex;

  assert(0 <= registerId);
  assert(registerId < MAX_REGISTERS);
  
  registerPolygonId = NRegisterPolygons[registerId];

  assert(NRegisterPolygons[registerId] < MAX_REGISTER_POLYGONS);
  NRegisterPolygons[registerId]++;
  
  nLoops = Polygon_nLoops[polygonId];
  RegisterPolygon_nLoops[registerId][registerPolygonId] 
    = nLoops;
  
  for (iLoop = 0; iLoop < nLoops; iLoop++) {
    int nVertexs = Polygon_nVertexs[polygonId][iLoop];
    
    RegisterPolygon_nVertexs[registerId][registerPolygonId][iLoop] 
      = nVertexs;
    
    for (iVertex = 0; iVertex < nVertexs; iVertex++) {
      double x = Polygon_vertexs[polygonId][iLoop][iVertex][0];
      double y = Polygon_vertexs[polygonId][iLoop][iVertex][1];
      double z = Polygon_vertexs[polygonId][iLoop][iVertex][2];
      
      RegisterPolygon_vertexs
	[registerId][registerPolygonId][iLoop][iVertex][0] = x;
      RegisterPolygon_vertexs
	[registerId][registerPolygonId][iLoop][iVertex][1] = y; 
      RegisterPolygon_vertexs
	[registerId][registerPolygonId][iLoop][iVertex][2] = z; 
    }
  }
}

static void AddRegisterPolygon 
(int srcRegisterId,
 int destRegisterId)
{
  int nPolygons;
  int offset;
  int iPolygon;
  int iLoop;
  int iVertex;

  assert(0 <= srcRegisterId);
  assert(srcRegisterId < MAX_REGISTERS);
  assert(0 <= destRegisterId);
  assert(destRegisterId < MAX_REGISTERS);

  nPolygons = NRegisterPolygons[srcRegisterId];
  offset = NRegisterPolygons[destRegisterId];

  NRegisterPolygons[destRegisterId] += nPolygons;

  for (iPolygon = 0; iPolygon < nPolygons; iPolygon++) {
    int nLoops = RegisterPolygon_nLoops[srcRegisterId][iPolygon];
    
    RegisterPolygon_nLoops[destRegisterId][iPolygon + offset] = nLoops;
    
    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs = RegisterPolygon_nVertexs[srcRegisterId][iPolygon][iLoop];
      
      RegisterPolygon_nVertexs
	[destRegisterId][iPolygon + offset][iLoop] = nVertexs;
      
      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	double x = RegisterPolygon_vertexs
	  [srcRegisterId][iPolygon][iLoop][iVertex][0];
	double y = RegisterPolygon_vertexs
	  [srcRegisterId][iPolygon][iLoop][iVertex][1];
	double z = RegisterPolygon_vertexs
	  [srcRegisterId][iPolygon][iLoop][iVertex][2];
	
	RegisterPolygon_vertexs
	  [destRegisterId][iPolygon + offset][iLoop][iVertex][0] = x;
	RegisterPolygon_vertexs
	  [destRegisterId][iPolygon + offset][iLoop][iVertex][1] = y;
	RegisterPolygon_vertexs
	  [destRegisterId][iPolygon + offset][iLoop][iVertex][2] = z;
      }
    }
  }
}

static void SetUVCoordinateOfRegisterPolygon 
(int registerId)
{
  assert(0 <= registerId);
  assert(registerId < MAX_REGISTERS);

  assert(0 < NRegisterPolygons[registerId]);

  {
    int v0 = 0;
    int v1 = 1;
    int v2 = RegisterPolygon_nVertexs[registerId][0][0] - 1;
    
    double x0 = RegisterPolygon_vertexs[registerId][0][0][v0][0];
    double y0 = RegisterPolygon_vertexs[registerId][0][0][v0][1];
    double z0 = RegisterPolygon_vertexs[registerId][0][0][v0][2];
    
    double x1 = RegisterPolygon_vertexs[registerId][0][0][v1][0];
    double y1 = RegisterPolygon_vertexs[registerId][0][0][v1][1];
    double z1 = RegisterPolygon_vertexs[registerId][0][0][v1][2];

    double x2 = RegisterPolygon_vertexs[registerId][0][0][v2][0];
    double y2 = RegisterPolygon_vertexs[registerId][0][0][v2][1];
    double z2 = RegisterPolygon_vertexs[registerId][0][0][v2][2];

    SetUVCoordinate 
      (x0, y0, z0,
       x1, y1, z1,
       x2, y2, z2);
  }
}

static void ReadRegisterPolygonFile 
(int registerId,
 const char *fileName)
{
  FILE *fp;
  int nPolygons;
  int iPolygon;
  int iLoop;
  int iVertex;
  
  assert(0 <= registerId);
  assert(registerId < MAX_REGISTERS);
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d",
	  &nPolygons);
  assert(0 <= nPolygons);

  NRegisterPolygons[registerId] = nPolygons;
  
  for (iPolygon = 0; iPolygon < nPolygons; iPolygon++) {
    int polygonId;
    int nLoops;
    
    fscanf (fp, "%d %d",
	    &polygonId,
	    &nLoops);
    assert(polygonId == iPolygon);
    assert(0 < nLoops);
    
    RegisterPolygon_nLoops[registerId][iPolygon] = nLoops;
    
    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs;

      fscanf (fp, "%d",
	      &nVertexs);
      assert(2 < nVertexs);
      
      RegisterPolygon_nVertexs[registerId][iPolygon][iLoop] = nVertexs;
      
      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	double u, v;
	double x, y, z;
	
	fscanf (fp, "%lf %lf",
		&u, &v);
	
	GetCoordinateOfUV (u, v,
			   &x, &y, &z);
	
	RegisterPolygon_vertexs
	  [registerId][iPolygon][iLoop][iVertex][0] = x;
	RegisterPolygon_vertexs
	  [registerId][iPolygon][iLoop][iVertex][1] = y;
	RegisterPolygon_vertexs
	  [registerId][iPolygon][iLoop][iVertex][2] = z;
      }
    }
  }

  fclose (fp);
}

static void WriteRegisterPolygonFile 
(int registerId,
 const char *fileName)
{
  FILE *fp;
  int nPolygons;
  int iPolygon;
  int iLoop;
  int iVertex;

  assert(0 <= registerId);
  assert(registerId < MAX_REGISTERS);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  nPolygons = NRegisterPolygons[registerId];
  fprintf (fp, "%d\n",
	   nPolygons);

  for (iPolygon = 0; iPolygon < nPolygons; iPolygon++) {
    int nLoops = RegisterPolygon_nLoops[registerId][iPolygon];
    
    fprintf (fp, "%d   %d\n",
	     iPolygon,
	     nLoops);

    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs = RegisterPolygon_nVertexs[registerId][iPolygon][iLoop];
      
      fprintf (fp, "  %d\n",
	       nVertexs);

      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	double x = RegisterPolygon_vertexs
	  [registerId][iPolygon][iLoop][iVertex][0];
	double y = RegisterPolygon_vertexs
	  [registerId][iPolygon][iLoop][iVertex][1];
	double z = RegisterPolygon_vertexs
	  [registerId][iPolygon][iLoop][iVertex][2];

	double u, v;

	GetUVOfCoordinate (x, y, z,
			   &u, &v);

	fprintf (fp, "  %22.15e %22.15e\n",
		 u, v);
      }
    }
  }

  fclose (fp);
}

static void WriteRegisterPolygonFileInReverseOrder 
(int registerId,
 const char *fileName)
{
  FILE *fp;
  int nPolygons;
  int iPolygon;
  int iLoop;
  int iVertex;

  assert(0 <= registerId);
  assert(registerId < MAX_REGISTERS);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  nPolygons = NRegisterPolygons[registerId];
  fprintf (fp, "%d\n",
	   nPolygons);

  for (iPolygon = 0; iPolygon < nPolygons; iPolygon++) {
    int nLoops = RegisterPolygon_nLoops[registerId][iPolygon];
    
    fprintf (fp, "%d   %d\n",
	     iPolygon,
	     nLoops);

    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs = RegisterPolygon_nVertexs[registerId][iPolygon][iLoop];
      
      fprintf (fp, "  %d\n",
	       nVertexs);

      for (iVertex = nVertexs - 1; 0 <= iVertex; iVertex--) {
	double x = RegisterPolygon_vertexs
	  [registerId][iPolygon][iLoop][iVertex][0];
	double y = RegisterPolygon_vertexs
	  [registerId][iPolygon][iLoop][iVertex][1];
	double z = RegisterPolygon_vertexs
	  [registerId][iPolygon][iLoop][iVertex][2];

	double u, v;

	GetUVOfCoordinate (x, y, z,
			   &u, &v);

	fprintf (fp, "  %22.15e %22.15e\n",
		 u, v);
      }
    }
  }

  fclose (fp);
}

static void WriteRegisterPolygonFileAsOutput 
(int registerId,
 const char *fileName)
{
  FILE *fp;
  int nPolygons;
  int iPolygon;
  int iLoop;
  int iVertex;

  assert(0 <= registerId);
  assert(registerId < MAX_REGISTERS);

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  nPolygons = NRegisterPolygons[registerId];
  fprintf (fp, "%d\n",
	   nPolygons);

  for (iPolygon = 0; iPolygon < nPolygons; iPolygon++) {
    int nLoops = RegisterPolygon_nLoops[registerId][iPolygon];
    
    fprintf (fp, "%d   %d\n",
	     iPolygon,
	     nLoops);

    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs = RegisterPolygon_nVertexs[registerId][iPolygon][iLoop];
      
      fprintf (fp, "  %d\n",
	       nVertexs);

      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	double x = RegisterPolygon_vertexs
	  [registerId][iPolygon][iLoop][iVertex][0];
	double y = RegisterPolygon_vertexs
	  [registerId][iPolygon][iLoop][iVertex][1];
	double z = RegisterPolygon_vertexs
	  [registerId][iPolygon][iLoop][iVertex][2];

	fprintf (fp, "  %22.15e %22.15e %22.15e\n",
		 x, y, z);
      }
    }
  }

  fclose (fp);
}

static void UnitePolygonOnPlane 
(int srcRegisterIdA, int srcRegisterIdB,
 int destRegisterId)
{
  char buf[2048];

  if (0 < NRegisterPolygons[srcRegisterIdA]
      && 0 < NRegisterPolygons[srcRegisterIdB]) {
    SetUVCoordinateOfRegisterPolygon (srcRegisterIdA);

    WriteRegisterPolygonFile (srcRegisterIdA, SET_POLYGON_FILE_NAME_A);
    WriteRegisterPolygonFile (srcRegisterIdB, SET_POLYGON_FILE_NAME_B);
    
    sprintf (buf, "%s %s %s add %s", 
	     SET2D_COMMAND,
	     SET_POLYGON_FILE_NAME_A,
	     SET_POLYGON_FILE_NAME_B,
	     SET_POLYGON_FILE_NAME_AB);
    system (buf);
    
    ReadRegisterPolygonFile (destRegisterId, SET_POLYGON_FILE_NAME_AB);

  } else if (0 < NRegisterPolygons[srcRegisterIdA]) {
    ClearRegisterPolygon (destRegisterId);
    AddRegisterPolygon (srcRegisterIdA, destRegisterId);

  } else if (0 < NRegisterPolygons[srcRegisterIdB]) {
    ClearRegisterPolygon (destRegisterId);
    AddRegisterPolygon (srcRegisterIdB, destRegisterId);

  } else {
    ClearRegisterPolygon (destRegisterId);
  }
}

static void SubtractPolygonOnPlane 
(int srcRegisterIdA, int srcRegisterIdB,
 int destRegisterId)
{
  char buf[2048];

  if (0 < NRegisterPolygons[srcRegisterIdA]
      && 0 < NRegisterPolygons[srcRegisterIdB]) {
    SetUVCoordinateOfRegisterPolygon (srcRegisterIdA);

    WriteRegisterPolygonFile (srcRegisterIdA, SET_POLYGON_FILE_NAME_A);
    WriteRegisterPolygonFileInReverseOrder 
      (srcRegisterIdB, SET_POLYGON_FILE_NAME_B);
    
    sprintf (buf, "%s %s %s sub %s", 
	     SET2D_COMMAND,
	     SET_POLYGON_FILE_NAME_A,
	     SET_POLYGON_FILE_NAME_B,
	     SET_POLYGON_FILE_NAME_AB);
    system (buf);
    
    ReadRegisterPolygonFile (destRegisterId, SET_POLYGON_FILE_NAME_AB);

  } else if (0 < NRegisterPolygons[srcRegisterIdA]) {
    ClearRegisterPolygon (destRegisterId);
    AddRegisterPolygon (srcRegisterIdA, destRegisterId);

  } else {
    ClearRegisterPolygon (destRegisterId);
  }
}

static void MergePolygonOnPlane (int planeId)
{
  int nPlanePolygons = Plane_nPolygons[planeId];
  int iPlanePolygon;

  assert(0 <= nPlanePolygons);

  if (1 < nPlanePolygons) {
    ClearRegisterPolygon (R0);
    ClearRegisterPolygon (R1);
    for (iPlanePolygon = 0; 
	 iPlanePolygon < nPlanePolygons; 
	 iPlanePolygon++) {
      int polygonId = Plane_polygonIds[planeId][iPlanePolygon];
      
      if (Polygon_isFrontSide[polygonId]) {
	if (Polygon_isA[polygonId]) {
	  AddPolygonIntoRegisterPolygon (R0, polygonId); 
	} else {
	  AddPolygonIntoRegisterPolygon (R1, polygonId); 
	}
      }
    }
    UnitePolygonOnPlane (R0, R1, R2);
    
    ClearRegisterPolygon (R3);
    ClearRegisterPolygon (R4);
    for (iPlanePolygon = 0; 
	 iPlanePolygon < nPlanePolygons; 
	 iPlanePolygon++) {
      int polygonId = Plane_polygonIds[planeId][iPlanePolygon];
      
      if (!Polygon_isFrontSide[polygonId]) {
	if (Polygon_isA[polygonId]) {
	  AddPolygonIntoRegisterPolygon (R3, polygonId); 
	} else {
	  AddPolygonIntoRegisterPolygon (R4, polygonId); 
	}
      }
    }
    UnitePolygonOnPlane (R3, R4, R5);
    
    SubtractPolygonOnPlane (R2, R5, R6);
    SubtractPolygonOnPlane (R5, R2, R7);
    
    AddRegisterPolygon (R6, R8);
    AddRegisterPolygon (R7, R8);

  } else {
    assert(nPlanePolygons == 1);

    for (iPlanePolygon = 0; 
	 iPlanePolygon < nPlanePolygons; 
	 iPlanePolygon++) {
      int polygonId = Plane_polygonIds[planeId][iPlanePolygon];
      
      AddPolygonIntoRegisterPolygon (R8, polygonId); 
    }
  }
}

static int NLoopVertexs;
static double LoopVertexs[MAX_POLYGON_VERTEXS][3];

static double LoopPlaneA;
static double LoopPlaneB;
static double LoopPlaneC;
static double LoopPlaneD;

static void TryToCutLoop 
(double x0, double y0, double z0,
 double x1, double y1, double z1)
{
  int iVertex;
  double intX, intY, intZ;
  int flag;
  int vertexIdFound;

  flag = GetIntersectionBetweenPlaneAndSeg 
    (LoopPlaneA, LoopPlaneB, LoopPlaneC, LoopPlaneD,
     x0, y0, z0,
     x1, y1, z1,
     &intX, &intY, &intZ);
  if (!flag) {
    return;
  }

  vertexIdFound = -1;
  for (iVertex = 0; iVertex < NLoopVertexs; iVertex++) {
    int v0 = iVertex;
    int v1 = (iVertex + 1) % NLoopVertexs;
    
    double sx0 = LoopVertexs[v0][0];
    double sy0 = LoopVertexs[v0][1];
    double sz0 = LoopVertexs[v0][2];
    
    double sx1 = LoopVertexs[v1][0];
    double sy1 = LoopVertexs[v1][1];
    double sz1 = LoopVertexs[v1][2];

    if (SegContainsPoint (sx0, sy0, sz0,
			  sx1, sy1, sz1,
			  intX, intY, intZ)) {
      if (!PointsAreSame (intX, intY, intZ,
			  sx0, sy0, sz0)
	  && !PointsAreSame (intX, intY, intZ,
			     sx1, sy1, sz1)) {
	vertexIdFound = iVertex;
	break;
      }
    }
  }

  if (0 <= vertexIdFound) {

    assert(NLoopVertexs < MAX_POLYGON_VERTEXS);

    for (iVertex = NLoopVertexs - 1; 
	 vertexIdFound < iVertex; 
	 iVertex--) {
      LoopVertexs[iVertex + 1][0] = LoopVertexs[iVertex][0];
      LoopVertexs[iVertex + 1][1] = LoopVertexs[iVertex][1];
      LoopVertexs[iVertex + 1][2] = LoopVertexs[iVertex][2];
    }

    LoopVertexs[vertexIdFound + 1][0] = intX;
    LoopVertexs[vertexIdFound + 1][1] = intY;
    LoopVertexs[vertexIdFound + 1][2] = intZ;

    NLoopVertexs++;
  }
}

static void InsertVertexIntoLoop (int srcRegisterId)
{
  int nPolygons;
  int iPolygon;
  int iLoop;
  int iVertex;

  nPolygons = NRegisterPolygons[srcRegisterId];

  for (iPolygon = 0; iPolygon < nPolygons; iPolygon++) {
    int nLoops = RegisterPolygon_nLoops[srcRegisterId][iPolygon];
    
    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs = RegisterPolygon_nVertexs[srcRegisterId][iPolygon][iLoop];
      
      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	int v0 = iVertex;
	int v1 = (iVertex + 1) % nVertexs;

	double x0 = RegisterPolygon_vertexs
	  [srcRegisterId][iPolygon][iLoop][v0][0];
	double y0 = RegisterPolygon_vertexs
	  [srcRegisterId][iPolygon][iLoop][v0][1];
	double z0 = RegisterPolygon_vertexs
	  [srcRegisterId][iPolygon][iLoop][v0][2];

	double x1 = RegisterPolygon_vertexs
	  [srcRegisterId][iPolygon][iLoop][v1][0];
	double y1 = RegisterPolygon_vertexs
	  [srcRegisterId][iPolygon][iLoop][v1][1];
	double z1 = RegisterPolygon_vertexs
	  [srcRegisterId][iPolygon][iLoop][v1][2];

	TryToCutLoop (x0, y0, z0,
		      x1, y1, z1);
      }
    }
  }
}

static void InsertVertexIntoPolygon 
(int srcRegisterId,
 int destRegisterId)
{
  int nPolygons;
  int iPolygon;
  int iLoop;
  int iVertex;

  nPolygons = NRegisterPolygons[srcRegisterId];

  NRegisterPolygons[destRegisterId] = nPolygons;

  for (iPolygon = 0; iPolygon < nPolygons; iPolygon++) {
    int nLoops = RegisterPolygon_nLoops[srcRegisterId][iPolygon];
    
    RegisterPolygon_nLoops[destRegisterId][iPolygon] = nLoops;

    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs = RegisterPolygon_nVertexs[srcRegisterId][iPolygon][iLoop];
      
      NLoopVertexs = nVertexs;

      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	double x = RegisterPolygon_vertexs
	  [srcRegisterId][iPolygon][iLoop][iVertex][0];
	double y = RegisterPolygon_vertexs
	  [srcRegisterId][iPolygon][iLoop][iVertex][1];
	double z = RegisterPolygon_vertexs
	  [srcRegisterId][iPolygon][iLoop][iVertex][2];

	LoopVertexs[iVertex][0] = x;
	LoopVertexs[iVertex][1] = y;
	LoopVertexs[iVertex][2] = z;
      }

      {
	double x0 = LoopVertexs[0][0];
	double y0 = LoopVertexs[0][1];
	double z0 = LoopVertexs[0][2];

	double x1 = LoopVertexs[1][0];
	double y1 = LoopVertexs[1][1];
	double z1 = LoopVertexs[1][2];

	double x2 = LoopVertexs[NLoopVertexs - 1][0];
	double y2 = LoopVertexs[NLoopVertexs - 1][1];
	double z2 = LoopVertexs[NLoopVertexs - 1][2];

	GetPlaneEquation 
	  (x0, y0, z0,
	   x1, y1, z1,
	   x2, y2, z2,
	   &LoopPlaneA, &LoopPlaneB, &LoopPlaneC, &LoopPlaneD);
      }

      InsertVertexIntoLoop (srcRegisterId);
      
      RegisterPolygon_nVertexs
	[destRegisterId][iPolygon][iLoop] = NLoopVertexs;
      
      for (iVertex = 0; iVertex < NLoopVertexs; iVertex++) {
	double x = LoopVertexs[iVertex][0];
	double y = LoopVertexs[iVertex][1];
	double z = LoopVertexs[iVertex][2];
	
	RegisterPolygon_vertexs
	  [destRegisterId][iPolygon][iLoop][iVertex][0] = x;
	RegisterPolygon_vertexs
	  [destRegisterId][iPolygon][iLoop][iVertex][1] = y;
	RegisterPolygon_vertexs
	  [destRegisterId][iPolygon][iLoop][iVertex][2] = z;
      }
    }
  }
}

void PerformCommand (const char *outputPolygonFileName)
{
  int iPlane;

  IdentifyPlane ();

  ClearRegisterPolygon (R8);
  for (iPlane = 0; iPlane < NPlanes; iPlane++) {
    MergePolygonOnPlane (iPlane);
  }

  InsertVertexIntoPolygon (R8, R9);
  WriteRegisterPolygonFileAsOutput (R9, outputPolygonFileName);
}



int main (int argc, char *argv[])
{
  if (argc != 4) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautoage_h_gm3d_attach_polygon  polygon_file_a  polygon_file_b  \n"
       "        polygon_file_OUT \n");
    exit (1);
  }

  ReadPolygonFile (argv[1], argv[2]);

  PerformCommand (argv[3]);

  return 0;
}
