/* advautoage_h_gm3d_extrude_polygon.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

void GetNormalizedVector 
(double x, double y, double z,
 double *nx_OUT, double *ny_OUT, double *nz_OUT)
{
  double length = sqrt (x * x + y * y + z * z);

  *nx_OUT = x / length;
  *ny_OUT = y / length;
  *nz_OUT = z / length;
}

void GetVectorProduct 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  *x_OUT = y0 * z1 - z0 * y1;
  *y_OUT = z0 * x1 - x0 * z1;
  *z_OUT = x0 * y1 - y0 * x1;
}



/********************************
         Polygon Data
********************************/

/* Polygon */

#define MAX_POLYGONS 1000
#define MAX_POLYGON_LOOPS 100
#define MAX_POLYGON_VERTEXS 100

int NInPolygons;
int InPolygon_nLoops[MAX_POLYGONS];
int InPolygon_nVertexs[MAX_POLYGONS][MAX_POLYGON_LOOPS];
double InPolygon_vertexs[MAX_POLYGONS][MAX_POLYGON_LOOPS][MAX_POLYGON_VERTEXS][3];

int NOutPolygons;
int OutPolygon_nLoops[MAX_POLYGONS];
int OutPolygon_nVertexs[MAX_POLYGONS][MAX_POLYGON_LOOPS];
double OutPolygon_vertexs[MAX_POLYGONS][MAX_POLYGON_LOOPS][MAX_POLYGON_VERTEXS][3];



void ReadInPolygonFile (const char *fileName)
{
  FILE *fp;
  int iPolygon;
  int iLoop;
  int iVertex;

  assert(NInPolygons == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NInPolygons);
  assert(0 < NInPolygons);
  assert(NInPolygons <= MAX_POLYGONS);
  
  for (iPolygon = 0; iPolygon < NInPolygons; iPolygon++) {
    int polygonId;
    int nLoops;
    
    fscanf (fp, "%d %d", 
	    &polygonId,
	    &nLoops);
    assert(polygonId == iPolygon);
    assert(0 < nLoops);
    
    InPolygon_nLoops[iPolygon] = nLoops;
    
    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs;

      fscanf (fp, "%d", 
	      &nVertexs);
      assert(2 < nVertexs);
      
      InPolygon_nVertexs[iPolygon][iLoop] = nVertexs;

      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	double x, y, z;

	fscanf (fp, "%lf %lf %lf", 
		&x, &y, &z);
	
	InPolygon_vertexs[iPolygon][iLoop][iVertex][0] = x;
	InPolygon_vertexs[iPolygon][iLoop][iVertex][1] = y; 
	InPolygon_vertexs[iPolygon][iLoop][iVertex][2] = z; 
      }
    }
  }

  fclose (fp);
}

int AddOutPolygon (void)
{
  int polygonId = NOutPolygons;

  NOutPolygons++;

  return polygonId;
}

int AddOutPolygonLoop (int polygonId)
{
  int loopId;

  assert(0 <= polygonId);
  assert(polygonId < NOutPolygons);

  loopId = OutPolygon_nLoops[polygonId];
  OutPolygon_nLoops[polygonId]++;

  return loopId;
}

void AddOutPolygonVertex 
(int polygonId, 
 int loopId,
 double x, double y, double z)
{
  int vertexId;

  assert(0 <= polygonId);
  assert(polygonId < NOutPolygons);
  assert(0 <= loopId);
  assert(loopId < OutPolygon_nLoops[polygonId]);

  vertexId = OutPolygon_nVertexs[polygonId][loopId];
  OutPolygon_nVertexs[polygonId][loopId]++;

  OutPolygon_vertexs[polygonId][loopId][vertexId][0] = x;
  OutPolygon_vertexs[polygonId][loopId][vertexId][1] = y;
  OutPolygon_vertexs[polygonId][loopId][vertexId][2] = z;
}

void WriteOutPolygonFile (const char *fileName)
{
  FILE *fp;
  int iPolygon;
  int iLoop;
  int iVertex;

  fp = fopen (fileName, "w");
  assert(fp != NULL);

  fprintf (fp, "%d\n",
	   NOutPolygons);

  for (iPolygon = 0; iPolygon < NOutPolygons; iPolygon++) {
    int nLoops = OutPolygon_nLoops[iPolygon];
    
    fprintf (fp, "%d   %d\n",
	     iPolygon,
	     nLoops);

    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs = OutPolygon_nVertexs[iPolygon][iLoop];
      
      fprintf (fp, "  %d\n",
	       nVertexs);

      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	double x = OutPolygon_vertexs[iPolygon][iLoop][iVertex][0];
	double y = OutPolygon_vertexs[iPolygon][iLoop][iVertex][1];
	double z = OutPolygon_vertexs[iPolygon][iLoop][iVertex][2];

	fprintf (fp, "  %22.15e %22.15e %22.15e\n",
		 x, y, z);
      }
    }
  }

  fclose (fp);
}



/************************
	 Command
*************************/

double Length;

double NormalX;
double NormalY;
double NormalZ;

static void SetNormal (void)
{
  double x0 = InPolygon_vertexs[0][0][0][0];
  double y0 = InPolygon_vertexs[0][0][0][1];
  double z0 = InPolygon_vertexs[0][0][0][2];
  
  double x1 = InPolygon_vertexs[0][0][1][0];
  double y1 = InPolygon_vertexs[0][0][1][1];
  double z1 = InPolygon_vertexs[0][0][1][2];
  
  double x2 = InPolygon_vertexs[0][0][2][0];
  double y2 = InPolygon_vertexs[0][0][2][1];
  double z2 = InPolygon_vertexs[0][0][2][2];

  double nx, ny, nz;

  GetVectorProduct (x1 - x0, y1 - y0, z1 - z0,
		    x2 - x0, y2 - y0, z2 - z0,
		    &nx, &ny, &nz);
  GetNormalizedVector (nx, ny, nz,
		       &nx, &ny, &nz);

  NormalX = nx;
  NormalY = ny;
  NormalZ = nz;
}

static void MakeTop (void)
{
  int nLoops = InPolygon_nLoops[0];
  int polygonId = AddOutPolygon ();
  int iLoop;
  int iVertex;
    
  for (iLoop = 0; iLoop < nLoops; iLoop++) {
    int nVertexs = InPolygon_nVertexs[0][iLoop];

    int loopId = AddOutPolygonLoop (polygonId);
    
    for (iVertex = 0; iVertex < nVertexs; iVertex++) {
      double x = InPolygon_vertexs[0][iLoop][iVertex][0];
      double y = InPolygon_vertexs[0][iLoop][iVertex][1];
      double z = InPolygon_vertexs[0][iLoop][iVertex][2];

      x += NormalX * Length;
      y += NormalY * Length;
      z += NormalZ * Length;

      AddOutPolygonVertex (polygonId, loopId,
			   x, y, z);
    }
  }
}

static void MakeButtom (void)
{
  int nLoops = InPolygon_nLoops[0];
  int polygonId = AddOutPolygon ();
  int iLoop;
  int iVertex;
    
  for (iLoop = 0; iLoop < nLoops; iLoop++) {
    int nVertexs = InPolygon_nVertexs[0][iLoop];

    int loopId = AddOutPolygonLoop (polygonId);
    
    for (iVertex = nVertexs - 1; 0 <= iVertex; iVertex--) {
      double x = InPolygon_vertexs[0][iLoop][iVertex][0];
      double y = InPolygon_vertexs[0][iLoop][iVertex][1];
      double z = InPolygon_vertexs[0][iLoop][iVertex][2];

      AddOutPolygonVertex (polygonId, loopId,
			   x, y, z);
    }
  }
}

static void MakeSide (void)
{
  int nLoops = InPolygon_nLoops[0];
  int iLoop;
  int iVertex;
    
  for (iLoop = 0; iLoop < nLoops; iLoop++) {
    int nVertexs = InPolygon_nVertexs[0][iLoop];

    for (iVertex = 0; iVertex < nVertexs; iVertex++) {
      int v0 = iVertex;
      int v1 = (iVertex + 1) % nVertexs;

      int polygonId = AddOutPolygon ();
      int loopId = AddOutPolygonLoop (polygonId);

      double x00 = InPolygon_vertexs[0][iLoop][v0][0];
      double y00 = InPolygon_vertexs[0][iLoop][v0][1];
      double z00 = InPolygon_vertexs[0][iLoop][v0][2];

      double x10 = InPolygon_vertexs[0][iLoop][v1][0];
      double y10 = InPolygon_vertexs[0][iLoop][v1][1];
      double z10 = InPolygon_vertexs[0][iLoop][v1][2];

      double x01 = x00 + NormalX * Length;
      double y01 = y00 + NormalY * Length;
      double z01 = z00 + NormalZ * Length;

      double x11 = x10 + NormalX * Length;
      double y11 = y10 + NormalY * Length;
      double z11 = z10 + NormalZ * Length;

      AddOutPolygonVertex (polygonId, loopId,
			   x00, y00, z00);
      AddOutPolygonVertex (polygonId, loopId,
			   x10, y10, z10);
      AddOutPolygonVertex (polygonId, loopId,
			   x11, y11, z11);
      AddOutPolygonVertex (polygonId, loopId,
			   x01, y01, z01);
    }
  }
}

void PerformCommand (void)
{
  assert(NInPolygons == 1);
  assert(0 < InPolygon_nLoops[0]);
  assert(3 <= InPolygon_nVertexs[0][0]);
  
  SetNormal ();
  MakeTop ();
  MakeButtom ();
  MakeSide ();
}



int main (int argc, char *argv[])
{
  if (argc != 4) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautoage_h_gm3d_extrude_polygon  polygon_file  \n"
       "          length \n"
       "          polygon_file_OUT \n");
    exit (1);
  }

  ReadInPolygonFile (argv[1]);
  Length = atof (argv[2]);

  PerformCommand ();

  WriteOutPolygonFile (argv[3]);

  return 0;
}
