/* advautoage_h_gm3d_face2triseg.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>



/************************
	 Basic
*************************/

void GetNormalizedVector 
(double x, double y, double z,
 double *nx_OUT, double *ny_OUT, double *nz_OUT)
{
  double length = sqrt (x * x + y * y + z * z);

  *nx_OUT = x / length;
  *ny_OUT = y / length;
  *nz_OUT = z / length;
}

void GetVectorProduct 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  *x_OUT = y0 * z1 - z0 * y1;
  *y_OUT = z0 * x1 - x0 * z1;
  *z_OUT = x0 * y1 - y0 * x1;
}

void GetPlaneEquation 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double x2, double y2, double z2,
 double *a_OUT, double *b_OUT, double *c_OUT, double *d_OUT)
{
  double nx, ny, nz;

  GetVectorProduct (x1 - x0, y1 - y0, z1 - z0,
		    x2 - x0, y2 - y0, z2 - z0,
		    &nx, &ny, &nz);
  GetNormalizedVector (nx, ny, nz,
		       &nx, &ny, &nz);
  *a_OUT = nx;
  *b_OUT = ny;
  *c_OUT = nz;
  *d_OUT = -(nx * x0 + ny * y0 + nz * z0);
}

void GetAnyPerpendicularDirection
(double x, double y, double z,
 double *nx0_OUT, double *ny0_OUT, double *nz0_OUT,
 double *nx1_OUT, double *ny1_OUT, double *nz1_OUT)
{
  double length;

  GetVectorProduct (x, y, z,
		    1, 0, 0,
		    nx0_OUT, ny0_OUT, nz0_OUT);
  length = sqrt (*nx0_OUT * *nx0_OUT
		 + *ny0_OUT * *ny0_OUT
		 + *nz0_OUT * *nz0_OUT);
  if (length == 0.0) {
    GetVectorProduct (x, y, z,
		      0, 1, 0,
		      nx0_OUT, ny0_OUT, nz0_OUT);
  }
  GetNormalizedVector (*nx0_OUT, *ny0_OUT, *nz0_OUT,
		       nx0_OUT, ny0_OUT, nz0_OUT);
  GetVectorProduct (x, y, z,
		    *nx0_OUT, *ny0_OUT, *nz0_OUT,
		    nx1_OUT, ny1_OUT, nz1_OUT);
  GetNormalizedVector (*nx1_OUT, *ny1_OUT, *nz1_OUT,
		       nx1_OUT, ny1_OUT, nz1_OUT);
}

void GetProjectedPointOnPlane
(double a, double b, double c, double d, 
 double x, double y, double z,
 double *px_OUT, double *py_OUT, double *pz_OUT)
{
  double value = a * x + b * y + c * z + d;
  
  *px_OUT = x - a * value;
  *py_OUT = y - b * value;
  *pz_OUT = z - c * value;
}



/********************************
         Face Data
********************************/

/* Point */

#define MAX_POINTS 10000

int NPoints;
double Point_coord[MAX_POINTS][3];



/* Edge */

#define MAX_EDGES 10000

int NEdges;
int Edge_pointId0[MAX_EDGES];
int Edge_pointId1[MAX_EDGES];



/* Face */

#define MAX_FACES 3000
#define MAX_FACE_LOOPS 100
#define MAX_FACE_VERTEXS 100

int NFaces;
int Face_nLoops[MAX_FACES];
int Face_nVertexs[MAX_FACES][MAX_FACE_LOOPS];
int Face_pointIds[MAX_FACES][MAX_FACE_LOOPS][MAX_FACE_VERTEXS];
int Face_edgeIds[MAX_FACES][MAX_FACE_LOOPS][MAX_FACE_VERTEXS];



void ReadPointFile (const char *fileName)
{
  FILE *fp;
  int iPoint;

  assert(NPoints == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NPoints);
  assert(0 < NPoints);
  assert(NPoints <= MAX_POINTS);
  
  for (iPoint = 0; iPoint < NPoints; iPoint++) {
    int pointId;
    double x, y, z;

    fscanf (fp, "%d %lf %lf %lf", 
	    &pointId,
	    &x, &y, &z);
    assert(pointId == iPoint);

    Point_coord[iPoint][0] = x;
    Point_coord[iPoint][1] = y; 
    Point_coord[iPoint][2] = z; 
  }

  fclose (fp);
}

void ReadEdgeFile (const char *fileName)
{
  FILE *fp;
  int iEdge;
  
  assert(0 < NPoints);
  assert(NEdges == 0);
  
  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NEdges);
  assert(0 < NEdges);
  assert(NEdges <= MAX_EDGES);
  
  for (iEdge = 0; iEdge < NEdges; iEdge++) {
    int edgeId;
    int pointId0, pointId1;

    fscanf (fp, "%d %d %d", 
	    &edgeId,
	    &pointId0, &pointId1);
    assert(edgeId == iEdge);
    assert(0 <= pointId0);
    assert(pointId0 < NPoints);
    assert(0 <= pointId1);
    assert(pointId1 < NPoints);

    Edge_pointId0[iEdge] = pointId0;
    Edge_pointId1[iEdge] = pointId1;
  }

  fclose (fp);
}

void ReadFaceFile (const char *fileName)
{
  FILE *fp;
  int iFace;
  int iLoop;
  int iVertex;

  assert(0 < NPoints);
  assert(0 < NEdges);
  assert(NFaces == 0);

  fp = fopen (fileName, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d", 
	  &NFaces);
  assert(0 < NFaces);
  assert(NFaces <= MAX_FACES);
  
  for (iFace = 0; iFace < NFaces; iFace++) {
    int faceId;
    int nLoops;
    
    fscanf (fp, "%d %d", 
	    &faceId,
	    &nLoops);
    assert(faceId == iFace);
    assert(0 < nLoops);
    
    Face_nLoops[iFace] = nLoops;
    
    for (iLoop = 0; iLoop < nLoops; iLoop++) {
      int nVertexs;

      fscanf (fp, "%d", 
	      &nVertexs);
      assert(2 < nVertexs);
      
      Face_nVertexs[iFace][iLoop] = nVertexs;
      
      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	int pointId;
	
	fscanf (fp, "%d", 
		&pointId);
	assert(0 <= pointId);
	assert(pointId < NPoints);
	
	Face_pointIds[iFace][iLoop][iVertex] = pointId;
      }

      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	int edgeId;

	fscanf (fp, "%d", 
		&edgeId);
	assert(0 <= edgeId);
	assert(edgeId < NEdges);
	
	Face_edgeIds[iFace][iLoop][iVertex] = edgeId;
      }

      for (iVertex = 0; iVertex < nVertexs; iVertex++) {
	int v0 = iVertex;
	int v1 = (iVertex + 1) % nVertexs;

	int p0 = Face_pointIds[iFace][iLoop][v0];
	int p1 = Face_pointIds[iFace][iLoop][v1];

	int edgeId = Face_edgeIds[iFace][iLoop][iVertex];
	int ep0 = Edge_pointId0[edgeId];
	int ep1 = Edge_pointId1[edgeId];

	assert((p0 == ep0 && p1 == ep1)
	       || (p0 == ep1 && p1 == ep0));
      }
    }
  }

  fclose (fp);
}



/************************
     Geometry  Data
*************************/

/* Segment */

#define MAX_SEGS 100000

int NSegs;
double Seg_coord0[MAX_SEGS][3];
double Seg_coord1[MAX_SEGS][3];



/* Triangle */

#define MAX_TRIS 100000

int NTris;
double Tri_coord0[MAX_TRIS][3];
double Tri_coord1[MAX_TRIS][3];
double Tri_coord2[MAX_TRIS][3];



int AddNewSeg 
(double x0, double y0, double z0,
 double x1, double y1, double z1)
{
  int segId = NSegs;

  assert(NSegs < MAX_SEGS);
  NSegs++;

  Seg_coord0[segId][0] = x0;
  Seg_coord0[segId][1] = y0;
  Seg_coord0[segId][2] = z0;

  Seg_coord1[segId][0] = x1;
  Seg_coord1[segId][1] = y1;
  Seg_coord1[segId][2] = z1;

  return segId;
}

int AddNewTri 
(double x0, double y0, double z0,
 double x1, double y1, double z1,
 double x2, double y2, double z2)
{
  int triId = NTris;

  assert(NTris < MAX_TRIS);
  NTris++;

  Tri_coord0[triId][0] = x0;
  Tri_coord0[triId][1] = y0;
  Tri_coord0[triId][2] = z0;

  Tri_coord1[triId][0] = x1;
  Tri_coord1[triId][1] = y1;
  Tri_coord1[triId][2] = z1;

  Tri_coord2[triId][0] = x2;
  Tri_coord2[triId][1] = y2;
  Tri_coord2[triId][2] = z2;

  return triId;
}



void WriteSegFile (const char *fileName)
{
  FILE *fp;
  int iSeg;

  assert(0 < NSegs);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  fprintf (fp, "%d\n", 
	   NSegs);
  
  for (iSeg = 0; iSeg < NSegs; iSeg++) {
    fprintf 
      (fp, "%d   %22.15e %22.15e %22.15e   %22.15e %22.15e %22.15e\n", 
       iSeg,
       Seg_coord0[iSeg][0], Seg_coord0[iSeg][1], Seg_coord0[iSeg][2],
       Seg_coord1[iSeg][0], Seg_coord1[iSeg][1], Seg_coord1[iSeg][2]);
  }

  fclose (fp);
}

void WriteTriFile (const char *fileName)
{
  FILE *fp;
  int iTri;

  assert(0 < NTris);

  fp = fopen (fileName, "w");
  assert(fp != NULL);
  
  fprintf (fp, "%d\n", 
	   NTris);
  
  for (iTri = 0; iTri < NTris; iTri++) {
    fprintf 
      (fp, 
       "%d   "
       "%22.15e %22.15e %22.15e   "
       "%22.15e %22.15e %22.15e   "
       "%22.15e %22.15e %22.15e\n", 
       iTri,
       Tri_coord0[iTri][0], Tri_coord0[iTri][1], Tri_coord0[iTri][2],
       Tri_coord1[iTri][0], Tri_coord1[iTri][1], Tri_coord1[iTri][2],
       Tri_coord2[iTri][0], Tri_coord2[iTri][1], Tri_coord2[iTri][2]);
  }

  fclose (fp);
}



/************************
	 Command
*************************/

static void MakeSeg (void)
{
  int iEdge;

  for (iEdge = 0; iEdge < NEdges; iEdge++) {
    int pointId0 = Edge_pointId0[iEdge];
    int pointId1 = Edge_pointId1[iEdge];
    
    double x0 = Point_coord[pointId0][0];
    double y0 = Point_coord[pointId0][1];
    double z0 = Point_coord[pointId0][2];
    
    double x1 = Point_coord[pointId1][0];
    double y1 = Point_coord[pointId1][1];
    double z1 = Point_coord[pointId1][2];
    
    AddNewSeg (x0, y0, z0, 
	       x1, y1, z1);
  }
}



#define SEG_FILE_NAME "___face2triseg__tessellate.gm_seg"
#define TRI_FILE_NAME "___face2triseg__tessellate.gm_tri"

#define TESSELLATE2D_COMMAND "advautoage_h_gm2d_tessellate"

static double Plane_a, Plane_b, Plane_c, Plane_d;
static double Plane_ox, Plane_oy, Plane_oz;
static double Plane_ux, Plane_uy, Plane_uz;
static double Plane_vx, Plane_vy, Plane_vz;

static void SetUVCoordinateOnFace (int faceId)
{
  const int outerLoopId = 0;
  int nVertexs = Face_nVertexs[faceId][outerLoopId];
  int pointId0, pointId1, pointId2;
  double x0, y0, z0;
  double x1, y1, z1;
  double x2, y2, z2;

  pointId0 = Face_pointIds[faceId][outerLoopId][0];
  pointId1 = Face_pointIds[faceId][outerLoopId][1];
  pointId2 = Face_pointIds[faceId][outerLoopId][nVertexs - 1];

  x0 = Point_coord[pointId0][0];
  y0 = Point_coord[pointId0][1];
  z0 = Point_coord[pointId0][2];
  
  x1 = Point_coord[pointId1][0];
  y1 = Point_coord[pointId1][1];
  z1 = Point_coord[pointId1][2];

  x2 = Point_coord[pointId2][0];
  y2 = Point_coord[pointId2][1];
  z2 = Point_coord[pointId2][2];

  GetPlaneEquation (x0, y0, z0,
		    x1, y1, z1,
		    x2, y2, z2,
		    &Plane_a, &Plane_b, &Plane_c, &Plane_d);
  GetProjectedPointOnPlane
    (Plane_a, Plane_b, Plane_c, Plane_d,
     x0, y0, z0, 
     &Plane_ox, &Plane_oy, &Plane_oz);
  GetAnyPerpendicularDirection
    (Plane_a, Plane_b, Plane_c,
     &Plane_ux, &Plane_uy, &Plane_uz,
     &Plane_vx, &Plane_vy, &Plane_vz);
  
}

static void GetUVOfCoordinateOnFace 
(double x, double y, double z,
 double *u_OUT, double *v_OUT)
{
  double u 
    = (x - Plane_ox) * Plane_ux
    + (y - Plane_oy) * Plane_uy
    + (z - Plane_oz) * Plane_uz;
  double v 
    = (x - Plane_ox) * Plane_vx
    + (y - Plane_oy) * Plane_vy
    + (z - Plane_oz) * Plane_vz;

  *u_OUT = u;
  *v_OUT = v;
}

static void GetCoordinateOfUVOnFace 
(double u, double v,
 double *x_OUT, double *y_OUT, double *z_OUT)
{
  double x = Plane_ox + Plane_ux * u + Plane_vx * v;
  double y = Plane_oy + Plane_uy * u + Plane_vy * v;
  double z = Plane_oz + Plane_uz * u + Plane_vz * v;

  *x_OUT = x;
  *y_OUT = y;
  *z_OUT = z;
}

static void WriteSeg2DFile (int faceId)
{
  int nLoops = Face_nLoops[faceId];
  FILE *fp;
  int iLoop;
  int iVertex;
  int count;

  fp = fopen (SEG_FILE_NAME, "w");
  assert(fp != NULL);

  count = 0;
  for (iLoop = 0; iLoop < nLoops; iLoop++) {
    int nVertexs = Face_nVertexs[faceId][iLoop];
    
    count += nVertexs;
  }

  fprintf (fp, "%d\n",
	   count);
  
  count = 0;
  for (iLoop = 0; iLoop < nLoops; iLoop++) {
    int nVertexs = Face_nVertexs[faceId][iLoop];
    
    for (iVertex = 0; iVertex < nVertexs; iVertex++) {
      int iv0 = iVertex;
      int iv1 = (iVertex + 1) % nVertexs;
      
      int pointId0 = Face_pointIds[faceId][iLoop][iv0];
      int pointId1 = Face_pointIds[faceId][iLoop][iv1];
      
      double x0 = Point_coord[pointId0][0];
      double y0 = Point_coord[pointId0][1];
      double z0 = Point_coord[pointId0][2];
      
      double x1 = Point_coord[pointId1][0];
      double y1 = Point_coord[pointId1][1];
      double z1 = Point_coord[pointId1][2];
      
      double u0, v0;
      double u1, v1;
      
      GetUVOfCoordinateOnFace (x0, y0, z0,
			       &u0, &v0);
      GetUVOfCoordinateOnFace (x1, y1, z1,
			       &u1, &v1);
      
      fprintf (fp, "%d   %22.15e %22.15e   %22.15e %22.15e\n",
	       count, 
	       u0, v0,
	       u1, v1);
      count++;
    }
  }
  
  fclose (fp);
}

static void ExecuteTessellation (void)
{
  char buf[2048];

  sprintf (buf, "%s %s %s", 
	   TESSELLATE2D_COMMAND,
	   SEG_FILE_NAME,
	   TRI_FILE_NAME);
  system (buf);
}

static void ReadTri2DFile (void)
{
  FILE *fp;
  int nTris;
  int iTri;
  
  fp = fopen (TRI_FILE_NAME, "r");
  assert(fp != NULL);
  
  fscanf (fp, "%d",
	  &nTris);
  
  for (iTri = 0; iTri < nTris; iTri++) {
    int triId;
    double u0, v0;
    double u1, v1;
    double u2, v2;
    double x0, y0, z0;
    double x1, y1, z1;
    double x2, y2, z2;
    
    fscanf (fp, "%d %lf %lf %lf %lf %lf %lf",
	    &triId,
	    &u0, &v0,
	    &u1, &v1,
	    &u2, &v2);
    assert(triId == iTri);

    GetCoordinateOfUVOnFace (u0, v0,
			     &x0, &y0, &z0);
    GetCoordinateOfUVOnFace (u1, v1,
			     &x1, &y1, &z1);
    GetCoordinateOfUVOnFace (u2, v2,
			     &x2, &y2, &z2);

    AddNewTri (x0, y0, z0,
	       x1, y1, z1,
	       x2, y2, z2);
  }

  fclose (fp);
}

static void MakeTri (void)
{
  int iFace;

  for (iFace = 0; iFace < NFaces; iFace++) {
    SetUVCoordinateOnFace (iFace);
    WriteSeg2DFile (iFace);
    ExecuteTessellation ();
    ReadTri2DFile ();
  }
}



void PerformCommand (void)
{
  MakeSeg ();
  MakeTri ();
}



int main (int argc, char *argv[])
{
  if (argc != 6) {
    fprintf 
      (stderr, 
       " Usage : \n"
       " advautoage_h_gm3d_face2triseg  point_file  \n"
       "        edge_file  face_file \n"
       "        seg_file_OUT \n"
       "        tri_file_OUT \n");
    exit (1);
  }

  ReadPointFile (argv[1]);
  ReadEdgeFile (argv[2]);
  ReadFaceFile (argv[3]);

  PerformCommand ();

  WriteSegFile (argv[4]);
  WriteTriFile (argv[5]);

  return 0;
}
